# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import logging as pylog

from qonos.common import timeutils
from qonos.common import utils
from qonos.openstack.common.notifier import api as notifier_api
from qonos.tests import utils as test_utils


class TestUtils(test_utils.BaseTestCase):

    def setUp(self):
        super(TestUtils, self).setUp()

    def tearDown(self):
        super(TestUtils, self).tearDown()
        self.stubs.UnsetAll()

    def test_serialize_datetimes(self):
        date_1 = datetime.datetime(2012, 5, 16, 15, 27, 36, 325355)
        date_2 = datetime.datetime(2013, 5, 16, 15, 27, 36, 325355)
        date_1_str = '2012-05-16T15:27:36Z'
        date_2_str = '2013-05-16T15:27:36Z'
        data = {'foo': date_1, 'bar': date_2}
        expected = {'foo': date_1_str, 'bar': date_2_str}
        utils.serialize_datetimes(data)
        self.assertEqual(data, expected)

    def test_serialize_datetimes_list(self):
        data = {
            'data': [
                {'foo': datetime.datetime(2012, 5, 16, 15, 27, 36, 325355)},
                {'foo': datetime.datetime(2013, 5, 16, 15, 27, 36, 325355)},
            ]
        }
        date_1_str = '2012-05-16T15:27:36Z'
        date_2_str = '2013-05-16T15:27:36Z'
        utils.serialize_datetimes(data)
        self.assertEqual(data['data'][0]['foo'], date_1_str)
        self.assertEqual(data['data'][1]['foo'], date_2_str)

    def test_serialize_datetimes_nested_dict(self):
        data = {
            'data': {
                'foo': {
                    'bar': datetime.datetime(2012, 5, 16, 15, 27, 36, 325355)
                }
            }
        }
        date_1_str = '2012-05-16T15:27:36Z'
        utils.serialize_datetimes(data)
        self.assertEqual(data['data']['foo']['bar'], date_1_str)

    def test_cron_string_to_datetime(self):
        minute = timeutils.utcnow().minute
        if minute == 0:
            minute = 59
        else:
            minute -= 1

        hour = timeutils.utcnow().hour
        if hour == 0:
            hour = 23
        else:
            hour -= 1
        next_run = utils.cron_string_to_next_datetime(minute=minute,
                                                      hour=hour)

        self.assertTrue(next_run > timeutils.utcnow())

    def test_cron_string_to_datetime_from_time(self):
        start_time = datetime.datetime(1980, 5, 16, 0, 0, 0, 0)
        next_run = utils.cron_string_to_next_datetime(minute=30,
                                                      hour=5,
                                                      start_time=start_time)
        expected = datetime.datetime(1980, 5, 16, 5, 30, 0, 0)

        self.assertTrue(next_run == expected)

    def test_generate_notification(self):

        notification = {}

        def fake_publisher_id(service, host=None):
            if not host:
                host = 'localhost'
            return "%s.%s" % (service, host)

        def fake_notify(context, publisher_id, event_type, priority, payload):
            notification['context'] = context
            notification['publisher_id'] = publisher_id
            notification['event_type'] = event_type
            notification['priority'] = priority
            notification['payload'] = payload

        self.stubs.Set(notifier_api, 'publisher_id', fake_publisher_id)
        self.stubs.Set(notifier_api, 'notify', fake_notify)
        payload = {'id': 'fake-id'}
        self.assertEqual(notification, {})
        self.config(host='localhost')
        utils.generate_notification(None, 'qonos.fake.event', payload)
        expected_notification = {
                                    'context': None,
                                    'publisher_id': 'qonos.localhost',
                                    'event_type': 'qonos.fake.event',
                                    'priority': 'INFO',
                                    'payload': {'id': 'fake-id'}
                                }

        self.assertEqual(notification, expected_notification)

    def test_get_qonos_open_file_log_handlers(self):

        class FakeStream(object):
            pass

        fake_stream = FakeStream()
        fake_stream.fileno = ''

        class FakeHandler(object):
            pass

        fake_handler = FakeHandler()
        fake_handler.stream = fake_stream

        class FakeLogger(object):
            pass

        fake_logger = FakeLogger()
        fake_logger.handlers = [None, fake_handler, FakeHandler()]

        def fake_get_logger(name):
            self.assertEqual(name, 'qonos')
            return fake_logger

        self.stubs.Set(pylog, 'getLogger', fake_get_logger)
        open_files = utils.get_qonos_open_file_log_handlers()
        self.assertEqual(open_files, [fake_stream])

    def test_log_warning_and_dismiss_exception(self):
        def mock_method():
            with utils.log_warning_and_dismiss_exception():
                raise Exception()
            return True
        self.assertTrue(mock_method())

    def test_log_warning_and_dismiss_exception_no_raise(self):
        def mock_method():
            with utils.log_warning_and_dismiss_exception():
                pass
            return True
        self.assertTrue(mock_method())



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

from oslo.config import cfg

CONF = cfg.CONF


netconf_opts = [
    cfg.StrOpt('host',
               default=socket.getfqdn(),
               help='Name of this node.  This can be an opaque identifier.  '
                    'It is not necessarily a hostname, FQDN, or IP address. '
                    'However, the node name must be valid within '
                    'an AMQP key, and if using ZeroMQ, a valid '
                    'hostname, FQDN, or IP address'),
]

CONF.register_opts(netconf_opts)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Defines interface for DB access
"""

import functools
import logging
import time

from oslo.config import cfg
import sqlalchemy
import sqlalchemy.orm as sa_orm
import sqlalchemy.sql as sa_sql

from qonos.common import exception
from qonos.common import timeutils
import qonos.db.db_utils as db_utils
from qonos.db.sqlalchemy import models
from qonos.openstack.common.gettextutils import _
import qonos.openstack.common.log as os_logging


_ENGINE = None
_MAKER = None
_MAX_RETRIES = None
_RETRY_INTERVAL = None
BASE = models.BASE
sa_logger = None
LOG = os_logging.getLogger(__name__)

db_opts = [
    cfg.IntOpt('sql_idle_timeout', default=3600),
    cfg.IntOpt('sql_max_retries', default=60),
    cfg.IntOpt('sql_retry_interval', default=1),
    cfg.BoolOpt('db_auto_create', default=True),
]

CONF = cfg.CONF
CONF.register_opts(db_opts)


def force_dict(func):
    """Ensure returned object is a dict or list of dicts."""

    def convert_list(a_list):
        to_return = []
        for object in a_list:
            if isinstance(object, list):
                to_return.append(convert_list(object))
            else:
                to_return.append(convert_object(object))
        return to_return

    def convert_object(object):
        if (isinstance(object, models.ModelBase) or
                isinstance(object, tuple)):
            to_return = dict(object)
        elif (isinstance(object, dict)):
            to_return = object
        else:
            raise ValueError()

        if 'parent' in to_return:
            del to_return['parent']

        if '_sa_instance_state' in to_return:
            del to_return['_sa_instance_state']

        for key in to_return:
            if isinstance(to_return[key], list):
                to_return[key] = convert_list(to_return[key])
            elif isinstance(to_return[key], models.ModelBase):
                to_return[key] = convert_object(to_return[key])

        return to_return

    def convert_output(output):
        if output is None:
            to_return = None
        elif isinstance(output, list):
            to_return = convert_list(output)
        else:
            to_return = convert_object(output)
        return to_return

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        output = func(*args, **kwargs)
        return convert_output(output)
    return wrapped


def ping_listener(dbapi_conn, connection_rec, connection_proxy):

    """
    Ensures that MySQL connections checked out of the
    pool are alive.

    Borrowed from:
    http://groups.google.com/group/sqlalchemy/msg/a4ce563d802c929f
    """

    try:
        dbapi_conn.cursor().execute('select 1')
    except dbapi_conn.OperationalError, ex:
        if ex.args[0] in (2006, 2013, 2014, 2045, 2055):
            msg = 'Got mysql server has gone away: %s' % ex
            LOG.warn(msg)
            raise sqlalchemy.exc.DisconnectionError(msg)
        else:
            raise


def configure_db():
    """
    Establish the database, create an engine if needed, and
    register the models.
    """
    global _ENGINE, sa_logger, _MAX_RETRIES, _RETRY_INTERVAL
    LOG.debug("Initializing DB")
    if not _ENGINE:
        sql_connection = CONF.sql_connection
        _MAX_RETRIES = CONF.sql_max_retries
        _RETRY_INTERVAL = CONF.sql_retry_interval
        connection_dict = sqlalchemy.engine.url.make_url(sql_connection)
        engine_args = {'pool_recycle': CONF.sql_idle_timeout,
                       'echo': False,
                       'convert_unicode': True
                       }

        try:
            _ENGINE = sqlalchemy.create_engine(sql_connection, **engine_args)

            if 'mysql' in connection_dict.drivername:
                sqlalchemy.event.listen(_ENGINE, 'checkout', ping_listener)

            _ENGINE.connect = wrap_db_error(_ENGINE.connect)
            _ENGINE.connect()
        except Exception, err:
            msg = _("Error configuring registry database with supplied "
                    "sql_connection '%(sql_connection)s'. "
                    "Got error:\n%(err)s") % locals()
            LOG.error(msg)
            raise

        sa_logger = logging.getLogger('sqlalchemy.engine')
        if CONF.debug:
            sa_logger.setLevel(logging.DEBUG)

        if CONF.db_auto_create:
            LOG.info('auto-creating qonos DB')
            models.register_models(_ENGINE)
        else:
            LOG.info('not auto-creating qonos DB')


def reset():
    models.unregister_models(_ENGINE)
    models.register_models(_ENGINE)


def get_session(autocommit=True, expire_on_commit=False):
    """Helper method to grab session."""
    global _MAKER
    if not _MAKER:
        assert _ENGINE
        _MAKER = sa_orm.sessionmaker(bind=_ENGINE,
                                     autocommit=autocommit,
                                     expire_on_commit=expire_on_commit)
    return _MAKER()


def is_db_connection_error(args):
    """Return True if error in connecting to db."""
    # NOTE(adam_g): This is currently MySQL specific and needs to be extended
    #               to support Postgres and others.
    conn_err_codes = ('2002', '2003', '2006')
    for err_code in conn_err_codes:
        if args.find(err_code) != -1:
            return True
    return False


def wrap_db_error(f):
    """Retry DB connection. Copied from nova and modified."""
    def _wrap(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except sqlalchemy.exc.OperationalError, e:
            if not is_db_connection_error(e.args[0]):
                raise

            remaining_attempts = _MAX_RETRIES
            while True:
                LOG.warning(_('SQL connection failed. %d attempts left.'),
                            remaining_attempts)
                remaining_attempts -= 1
                time.sleep(_RETRY_INTERVAL)
                try:
                    return f(*args, **kwargs)
                except sqlalchemy.exc.OperationalError, e:
                    if (remaining_attempts == 0 or
                            not is_db_connection_error(e.args[0])):
                        raise
                except sqlalchemy.exc.DBAPIError:
                    raise
        except sqlalchemy.exc.DBAPIError:
            raise
    _wrap.func_name = f.func_name
    return _wrap


#################### Schedule methods


@force_dict
def schedule_create(schedule_values):
    db_utils.validate_schedule_values(schedule_values)
    # make a copy so we can remove 'schedule_metadata'
    # without affecting the caller
    values = schedule_values.copy()
    metadata = None
    session = get_session()
    schedule_ref = models.Schedule()

    if 'schedule_metadata' in values:
        metadata = values['schedule_metadata']
        _set_schedule_metadata(schedule_ref, metadata)
        del values['schedule_metadata']

    schedule_ref.update(values)
    schedule_ref.save(session=session)

    return _schedule_get_by_id(schedule_ref['id'])


def paginate_query(query, model, sort_keys, limit=None, marker=None):
    """
    Returns a query with sorting and(or) pagination criteria added.

    Pagination works by requiring a unique sort_key, specified by sort_keys.
    (If sort_keys is not unique, then we risk looping through values.)
    We use the last row in the previous page as the 'marker' for pagination.
    So we must return values that follow the passed marker in the order.
    With a single-valued sort_key, this would be easy: sort_key > X.
    With a compound-values sort_key, (k1, k2, k3) we must do this to repeat
    the lexicographical ordering:
    (k1 > X1) or (k1 == X1 && k2 > X2) or (k1 == X1 && k2 == X2 && k3 > X3)

    Typically, the id of the last row is used as the client-facing pagination
    marker, then the actual marker object must be fetched from the db and
    passed in to us as marker.

    :param query: the query object to which we should add paging/sorting
    :param model: the ORM model class
    :param sort_keys: array of attributes by which results should be sorted
    :param limit: maximum number of items to return
    :param marker: the last item of the previous page; we returns the next
                    results after this value.

    :rtype: sqlalchemy.orm.query.Query
    :return: The query with sorting and(or) pagination added.
    """
    # Note(nikhil): Curently pagination does not support sort directions. By
    # default the items are sorted in ascending order.
    for sort_key in sort_keys:
        try:
            sort_key_attr = getattr(model, sort_key)
        except AttributeError:
            raise exception.InvalidSortKey()
        query = query.order_by(sqlalchemy.asc(sort_key_attr))

    if marker is not None:
        marker_values = []
        for sort_key in sort_keys:
            v = getattr(marker, sort_key)
            marker_values.append(v)

    # Note(nikhil): the underlying code only supports asc order of sort_dir
    #at the moment. However, more than one sort_keys could be supplied.
        criteria_list = []
        for i in xrange(0, len(sort_keys)):
            crit_attrs = []
            for j in xrange(0, i):
                model_attr = getattr(model, sort_keys[j])
                crit_attrs.append((model_attr == marker_values[j]))
            model_attr = getattr(model, sort_keys[i])
            crit_attrs.append((model_attr > marker_values[i]))
            criteria = sa_sql.and_(*crit_attrs)
            criteria_list.append(criteria)

        f = sa_sql.or_(*criteria_list)
        query = query.filter(f)

    if limit is not None:
        query = query.limit(limit)

    return query


@force_dict
def schedule_get_all(filter_args={}):
    session = get_session()
    query = session.query(models.Schedule)\
                   .options(sa_orm.joinedload_all(
                            models.Schedule.schedule_metadata))
    SCHEDULE_BASE_FILTERS = ['next_run_after', 'next_run_before', 'tenant',
                    'limit', 'marker', 'action']

    if 'next_run_after' in filter_args:
        query = query.filter(
            models.Schedule.next_run >= filter_args['next_run_after'])
        filter_args.pop('next_run_after')

    if 'next_run_before' in filter_args:
        query = query.filter(
            models.Schedule.next_run <= filter_args['next_run_before'])

    if filter_args.get('tenant') is not None:
        query = query.filter(
                models.Schedule.tenant == filter_args['tenant'])

    if filter_args.get('action') is not None:
        query = query.filter(
                models.Schedule.action == filter_args['action'])

    for filter_key in filter_args.keys():
        if filter_key not in SCHEDULE_BASE_FILTERS:
            query = query.filter(models.Schedule.schedule_metadata.any(
                        key=filter_key, value=filter_args[filter_key]))

    marker_schedule = None
    if filter_args.get('marker') is not None:
        marker_schedule = _schedule_get_by_id(filter_args['marker'])

    query = paginate_query(query, models.Schedule, ['id'],
                           limit=filter_args.get('limit'),
                           marker=marker_schedule)

    return query.all()


def _schedule_get_by_id(schedule_id, session=None):
    session = session or get_session()
    try:
        schedule = session.query(models.Schedule)\
                          .options(sa_orm.joinedload_all('schedule_metadata'))\
                          .filter_by(id=schedule_id)\
                          .one()
    except sa_orm.exc.NoResultFound:
        raise exception.NotFound()

    return schedule


@force_dict
def schedule_get_by_id(schedule_id):
    return _schedule_get_by_id(schedule_id)


@force_dict
def schedule_update(schedule_id, schedule_values):
    # make a copy so we can remove 'schedule_metadata'
    # without affecting the caller
    values = schedule_values.copy()
    session = get_session()
    schedule_ref = _schedule_get_by_id(schedule_id, session)

    if 'schedule_metadata' in values:
        metadata = values.pop('schedule_metadata')
        _schedule_metadata_update_in_place(schedule_ref, metadata)

    schedule_ref.update(values)
    schedule_ref.save(session=session)
    return _schedule_get_by_id(schedule_id)


def schedule_test_and_set_next_run(schedule_id, expected_next_run, next_run):
    session = get_session()
    if expected_next_run:
        query = session.query(models.Schedule).filter_by(id=schedule_id)\
                       .filter_by(next_run=expected_next_run)\
                       .update(dict(next_run=next_run))
    else:
        query = session.query(models.Schedule).filter_by(id=schedule_id)\
                       .update(dict(next_run=next_run))

    if not query:
        raise exception.NotFound()


def _schedule_metadata_update_in_place(schedule, metadata):
    new_meta = {}
    for item in metadata:
        new_meta[item['key']] = item['value']
    to_delete = []
    for meta in schedule.schedule_metadata:
        if not meta['key'] in new_meta:
            to_delete.append(meta)
        else:
            meta['value'] = new_meta[meta['key']]
            del new_meta[meta['key']]

    for key, value in new_meta.iteritems():
        meta_ref = models.ScheduleMetadata()
        meta_ref.key = key
        meta_ref.value = value
        schedule.schedule_metadata.append(meta_ref)

    for meta in to_delete:
        schedule['schedule_metadata'].remove(meta)


def schedule_delete(schedule_id):
    session = get_session()
    schedule_ref = _schedule_get_by_id(schedule_id, session)
    schedule_ref.delete(session=session)


def _set_schedule_metadata(schedule_ref, metadata):
    for metadatum in metadata:
        metadata_ref = models.ScheduleMetadata()
        metadata_ref.update(metadatum)
        schedule_ref.schedule_metadata.append(metadata_ref)

#################### Schedule Metadata methods


@force_dict
def schedule_meta_create(schedule_id, values):
    session = get_session()
    _schedule_get_by_id(schedule_id, session)
    meta_ref = models.ScheduleMetadata()
    values['schedule_id'] = schedule_id
    meta_ref.update(values)

    try:
        meta_ref.save(session=session)
    except sqlalchemy.exc.IntegrityError:
        raise exception.Duplicate()

    return _schedule_meta_get(schedule_id, values['key'])


@force_dict
def schedule_meta_get_all(schedule_id):
    session = get_session()
    _schedule_get_by_id(schedule_id, session)
    query = session.query(models.ScheduleMetadata)\
                   .filter_by(schedule_id=schedule_id)

    return query.all()


def _schedule_meta_get(schedule_id, key, session=None):
    session = session or get_session()
    try:
        _schedule_get_by_id(schedule_id, session)
    except exception.NotFound:
        msg = _('Schedule %s could not be found') % schedule_id
        raise exception.NotFound(message=msg)
    try:
        meta = session.query(models.ScheduleMetadata)\
                      .filter_by(schedule_id=schedule_id)\
                      .filter_by(key=key)\
                      .one()
    except sa_orm.exc.NoResultFound:
        raise exception.NotFound()

    return meta


def _schedule_meta_update(schedule_id, key, values):
    session = get_session()
    _schedule_get_by_id(schedule_id, session)
    meta_ref = _schedule_meta_get(schedule_id, key, session)
    meta_ref.update(values)
    meta_ref.save(session=session)
    return meta_ref


@force_dict
def schedule_metadata_update(schedule_id, values):
    session = get_session()
    schedule = _schedule_get_by_id(schedule_id, session)
    _schedule_metadata_update_in_place(schedule, values)

    schedule.save(session=session)

    return schedule_meta_get_all(schedule_id)


def schedule_meta_delete(schedule_id, key):
    session = get_session()
    _schedule_get_by_id(schedule_id, session)
    meta_ref = _schedule_meta_get(schedule_id, key, session)
    meta_ref.delete(session=session)


##################### Worker methods


@force_dict
def worker_get_all(params={}):
    session = get_session()
    query = session.query(models.Worker)

    marker_worker = None
    if params.get('marker') is not None:
        marker_worker = _worker_get_by_id(params['marker'])

    query = paginate_query(query, models.Worker, ['id'],
                           limit=params.get('limit'), marker=marker_worker)

    return query.all()


@force_dict
def worker_create(values):
    session = get_session()
    worker_ref = models.Worker()
    worker_ref.update(values)
    worker_ref.save(session=session)

    return _worker_get_by_id(worker_ref['id'])


def _worker_get_by_id(worker_id):
    session = get_session()
    query = session.query(models.Worker).filter_by(id=worker_id)

    try:
        worker = query.one()
    except sa_orm.exc.NoResultFound:
        raise exception.NotFound

    return worker


@force_dict
def worker_get_by_id(worker_id):
    return _worker_get_by_id(worker_id)


def worker_delete(worker_id):
    session = get_session()

    with session.begin():
        worker = _worker_get_by_id(worker_id)
        worker.delete(session=session)


#################### Job methods


@force_dict
def job_create(job_values):
    db_utils.validate_job_values(job_values)
    values = job_values.copy()
    session = get_session()
    job_ref = models.Job()

    if 'job_metadata' in values:
        metadata = values['job_metadata']
        _set_job_metadata(job_ref, metadata)
        del values['job_metadata']

    job_ref.update(values)
    job_ref.save(session=session)

    return _job_get_by_id(job_ref['id'])


@force_dict
def job_get_all(params={}):
    session = get_session()
    query = session.query(models.Job)\
                   .options(sa_orm.subqueryload('job_metadata'))

    marker_job = None
    if params.get('marker') is not None:
        marker_job = _job_get_by_id(params['marker'])

    query = paginate_query(query, models.Job, ['id'],
                           limit=params.get('limit'), marker=marker_job)

    return query.all()


def _job_get_by_id(job_id, session=None):
    session = session or get_session()
    try:
        job = session.query(models.Job)\
                     .options(sa_orm.subqueryload('job_metadata'))\
                     .filter_by(id=job_id)\
                     .one()
    except sa_orm.exc.NoResultFound:
        raise exception.NotFound()

    return job


@force_dict
def job_get_by_id(job_id):
    return _job_get_by_id(job_id)


def job_updated_at_get_by_id(job_id):
    return _job_get_by_id(job_id)['updated_at']


@force_dict
def job_get_and_assign_next_by_action(action, worker_id, max_retry,
                                      new_timeout):
    """Get the next available job for the given action and assign it
    to the worker for worker_id."""
    now = timeutils.utcnow()
    session = get_session()

    job_ref = _job_get_next_by_action(session, now, action, max_retry)

    if not job_ref:
        return None

    # Make sure the job has not changed unexpectedly since
    # retrieving it
    try:
        query = session.query(models.Job).filter_by(id=job_ref['id'])\
                       .filter_by(updated_at=job_ref['updated_at'])\
                       .update({'worker_id': worker_id,
                                'timeout': new_timeout,
                                'retry_count': job_ref['retry_count'] + 1})
    except sa_orm.exc.NoResultFound:
        #In case the job was deleted during assignment return nothing
        return None

    if not query:
        return None

    return _job_get_by_id(job_ref['id'])


def _job_get_next_by_action(session, now, action, max_retry):
    job_ref = session.query(models.Job)\
        .options(sa_orm.subqueryload('job_metadata'))\
        .filter_by(action=action)\
        .filter(models.Job.retry_count < max_retry)\
        .filter(models.Job.hard_timeout > now)\
        .filter(sa_sql.or_(~models.Job.status.in_(['DONE', 'CANCELLED']),
                           models.Job.status == None))\
        .filter(sa_sql.or_(models.Job.worker_id == None,
                           models.Job.timeout <= now))\
        .order_by(models.Job.created_at.asc())\
        .first()
    return job_ref


def _jobs_cleanup_hard_timed_out():
    """Find all jobs with hard_timeout values which have passed
    and delete them, logging the timeout / failure as appropriate"""
    now = timeutils.utcnow()
    session = get_session()
    num_del = session.query(models.Job)\
        .filter(models.Job.hard_timeout <= now)\
        .delete()
    session.flush()
    return num_del


@force_dict
def job_update(job_id, job_values):
    # make a copy so we can remove 'job_metadata'
    # without affecting the caller
    values = job_values.copy()
    session = get_session()
    job_ref = _job_get_by_id(job_id, session)

    if 'job_metadata' in values:
        metadata = values.pop('job_metadata')
        _job_metadata_update_in_place(job_ref, metadata)

    job_ref.update(values)
    job_ref.save(session=session)
    return _job_get_by_id(job_id)


def _job_metadata_update_in_place(job, metadata):
    new_meta = {}
    for item in metadata:
        new_meta[item['key']] = item['value']
    to_delete = []
    for meta in job.job_metadata:
        if not meta['key'] in new_meta:
            to_delete.append(meta)
        else:
            meta['value'] = new_meta[meta['key']]
            del new_meta[meta['key']]

    for key, value in new_meta.iteritems():
        meta_ref = models.JobMetadata()
        meta_ref.key = key
        meta_ref.value = value
        job.job_metadata.append(meta_ref)

    for meta in to_delete:
        job['job_metadata'].remove(meta)


def job_delete(job_id):
    session = get_session()
    job_ref = _job_get_by_id(job_id)
    job_ref.delete(session=session)


def _set_job_metadata(job_ref, metadata):
    for metadatum in metadata:
        metadata_ref = models.JobMetadata()
        metadata_ref.update(metadatum)
        job_ref.job_metadata.append(metadata_ref)


@force_dict
def job_meta_create(job_id, values):
    values['job_id'] = job_id
    session = get_session()
    meta_ref = models.JobMetadata()
    meta_ref.update(values)

    try:
        meta_ref.save(session=session)
    except sqlalchemy.exc.IntegrityError:
        raise exception.Duplicate()

    return _job_meta_get_by_id(meta_ref['id'])


def _job_meta_get_by_id(meta_id):
    session = get_session()
    try:
        meta = session.query(models.JobMetadata)\
                      .filter_by(id=meta_id)\
                      .one()
    except sa_orm.exc.NoResultFound:
        raise exception.NotFound()

    return meta


def _job_meta_get_all_by_job_id(job_id):
    session = get_session()
    try:
        meta = session.query(models.JobMetadata)\
                      .filter_by(job_id=job_id)\
                      .all()
    except sa_orm.exc.NoResultFound:
        raise exception.NotFound()

    return meta


def _job_meta_get(job_id, key):
    session = get_session()
    try:
        meta = session.query(models.JobMetadata)\
                      .filter_by(job_id=job_id)\
                      .filter_by(key=key)\
                      .one()
    except sa_orm.exc.NoResultFound:
        raise exception.NotFound()

    return meta


@force_dict
def job_meta_get_all_by_job_id(job_id):
    return _job_meta_get_all_by_job_id(job_id)


@force_dict
def job_metadata_update(job_id, values):
    session = get_session()
    job = _job_get_by_id(job_id, session)
    _job_metadata_update_in_place(job, values)

    job.save(session=session)

    return job_meta_get_all_by_job_id(job_id)


##################### Job fault methods

def job_fault_latest_for_job_id(job_id):
    session = get_session()
    try:
        job_fault = session.query(models.JobFault)\
            .filter_by(job_id=job_id)\
            .order_by(models.JobFault.created_at.desc())\
            .first()
        return job_fault
    except sa_orm.exc.NoResultFound:
        return None


@force_dict
def job_fault_create(values):
    session = get_session()
    job_fault_ref = models.JobFault()
    job_fault_ref.update(values)

    job_fault_ref.save(session=session)

    return job_fault_ref



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid
import webob.exc

from qonos.api.v1 import schedule_metadata
from qonos.db.simple import api as db_api
from qonos.tests.unit import utils as unit_utils
from qonos.tests import utils as test_utils


class TestScheduleMetadataApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestScheduleMetadataApi, self).setUp()
        self.controller = schedule_metadata.\
            ScheduleMetadataController(db_api=db_api)
        self._create_schedules()

    def tearDown(self):
        super(TestScheduleMetadataApi, self).tearDown()
        db_api.reset()

    def _create_schedules(self):
        fixture = {
            'id': unit_utils.SCHEDULE_UUID1,
            'tenant': unit_utils.TENANT1,
            'action': 'snapshot',
            'minute': '30',
            'hour': '2',
        }
        self.schedule_1 = db_api.schedule_create(fixture)
        fixture = {
            'id': unit_utils.SCHEDULE_UUID2,
            'tenant': unit_utils.TENANT2,
            'action': 'snapshot',
            'minute': '30',
            'hour': '2',
        }
        self.schedule_2 = db_api.schedule_create(fixture)

    def test_list_meta(self):
        request = unit_utils.get_fake_request(method='POST')
        fixture = {'metadata': {'key1': 'value1', 'key2': 'value2'}}
        actual = self.controller.update(request, self.schedule_1['id'],
                                        fixture)
        self.assertEqual(actual, fixture)

    def test_update_metadata(self):
        request = unit_utils.get_fake_request(method='PUT')
        expected = {'metadata': {'key1': 'value1'}}
        actual = self.controller.update(request, self.schedule_1['id'],
                                        expected)

        self.assertEqual(expected, actual)

    def test_update_metadata_empty(self):
        request = unit_utils.get_fake_request(method='PUT')
        expected = {'metadata': {}}
        actual = self.controller.update(request, self.schedule_1['id'],
                                        expected)

        self.assertEqual(expected, actual)

    def test_update_meta_schedule_not_found(self):
        request = unit_utils.get_fake_request(method='PUT')
        schedule_id = uuid.uuid4()
        fixture = {'metadata': {'key1': 'value1'}}
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.update,
                          request, schedule_id, fixture)



# Copyright 2013 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Fakes For Worker tests.
"""

import datetime

from qonos.common import timeutils

WORKER_ID = '11111111-1111-1111-1111-11111111'
JOB_ID = '22222222-2222-2222-2222-22222222'
SCHEDULE_ID = '33333333-3333-3333-3333-33333333'
TENANT = '44444444-4444-4444-4444-44444444'
INSTANCE_ID = '55555555-5555-5555-5555-55555555'

BASE_TIME = timeutils.utcnow()
TIMEOUT = BASE_TIME + datetime.timedelta(hours=1)
HARD_TIMEOUT = BASE_TIME + datetime.timedelta(hours=2)

JOB = {
    'job': {
        'id': JOB_ID,
        'created_at': BASE_TIME,
        'modified_at': BASE_TIME,
        'schedule_id': SCHEDULE_ID,
        'tenant': TENANT,
        'worker_id': WORKER_ID,
        'status': 'QUEUED',
        'action': 'snapshot',
        'retry_count': 3,
        'timeout': TIMEOUT,
        'hard_timeout': HARD_TIMEOUT,
        'metadata': {
            'instance_id': INSTANCE_ID,
            },
    }
}

JOB_NONE = {
    'job': None
}

WORKER_HOST = 'snapshotters'
WORKER_NAME = 'snapshot_1'

WORKER = {
    'id': WORKER_ID,
    'created_at': BASE_TIME,
    'modified_at': BASE_TIME,
    'host': WORKER_HOST,
}



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2013 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
SQLAlchemy models for glance data
"""

from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import ForeignKey, DateTime, Text
from sqlalchemy.orm import relationship, backref, object_mapper
from sqlalchemy import UniqueConstraint

from qonos.common import timeutils
import qonos.db.sqlalchemy.api
from qonos.openstack.common import uuidutils

BASE = declarative_base()


class ModelBase(object):
    """Base class for Nova and Glance Models."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __table_initialized__ = False

    created_at = Column(DateTime, default=timeutils.utcnow,
                        nullable=False)
    updated_at = Column(DateTime, default=timeutils.utcnow,
                        nullable=False, onupdate=timeutils.utcnow)
    id = Column(String(36), primary_key=True, default=uuidutils.generate_uuid)

    def save(self, session=None):
        """Save this object."""
        session = session or qonos.db.sqlalchemy.api.get_session()
        session.add(self)
        session.flush()

    def delete(self, session=None):
        """Delete this object."""
        session = session or qonos.db.sqlalchemy.api.get_session()
        session.delete(self)
        session.flush()

    def update(self, values):
        """dict.update() behaviour."""
        for k, v in values.iteritems():
            self[k] = v

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __iter__(self):
        self._i = iter(object_mapper(self).columns)
        return self

    def next(self):
        n = self._i.next().name
        return n, getattr(self, n)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def to_dict(self):
        return self.__dict__.copy()


class Schedule(BASE, ModelBase):
    """Represents a schedule in the datastore."""
    __tablename__ = 'schedules'

    tenant = Column(String(255), nullable=False)
    action = Column(String(255), nullable=False)
    minute = Column(Integer, nullable=True)
    hour = Column(Integer, nullable=True)
    day_of_month = Column(Integer, nullable=True)
    month = Column(Integer, nullable=True)
    day_of_week = Column(Integer, nullable=True)
    last_scheduled = Column(DateTime, nullable=True)
    next_run = Column(DateTime, nullable=True)


class ScheduleMetadata(BASE, ModelBase):
    """Represents metadata of a schedule in the datastore."""
    __tablename__ = 'schedule_metadata'
    __table_args__ = (UniqueConstraint('schedule_id', 'key'), {})

    schedule_id = Column(String(36),
                         ForeignKey('schedules.id'), nullable=False)
    key = Column(String(255), nullable=False)
    value = Column(Text, nullable=False)
    parent = relationship(Schedule, backref=backref('schedule_metadata',
                                                    cascade='all,delete,'
                                                            'delete-orphan'))


class Worker(BASE, ModelBase):
    """Represents a worker in the datastore."""
    __tablename__ = 'workers'

    host = Column(String(255), nullable=False)
    process_id = Column(Integer, nullable=True)


class Job(BASE, ModelBase):
    """Represents a job in the datastore."""
    __tablename__ = 'jobs'

    schedule_id = Column(String(36))
    tenant = Column(String(255), nullable=False)
    worker_id = Column(String(36), nullable=True)
    status = Column(String(255), nullable=True)
    action = Column(String(255), nullable=False)
    retry_count = Column(Integer, nullable=False, default=0)
    timeout = Column(DateTime, nullable=False)
    hard_timeout = Column(DateTime, nullable=False)


class JobMetadata(BASE, ModelBase):
    """Represents job metadata in the datastore."""
    __tablename__ = 'job_metadata'
    __table_args__ = (UniqueConstraint('job_id', 'key'), {})

    job_id = Column(String(36), ForeignKey('jobs.id'), nullable=False)
    key = Column(String(255), nullable=False)
    value = Column(Text, nullable=False)
    parent = relationship(Job, backref=backref('job_metadata',
                                               cascade='all,delete,'
                                                       'delete-orphan'))


class JobFault(BASE, ModelBase):
    """Represents a job fault in the datastore."""
    __tablename__ = 'job_faults'

    job_id = Column(String(36), nullable=False)
    schedule_id = Column(String(36), nullable=False)
    tenant = Column(String(255), nullable=False)
    worker_id = Column(String(36), nullable=False)
    action = Column(String(255), nullable=False)
    message = Column(String(255), nullable=True)
    job_metadata = Column(Text, nullable=True)


def register_models(engine):
    """
    Creates database tables for all models with the given engine.
    """
    models = (Schedule, ScheduleMetadata, Worker, Job, JobFault)
    for model in models:
        model.metadata.create_all(engine)


def unregister_models(engine):
    """
    Drops database tables for all models with the given engine.
    """
    models = (Schedule, ScheduleMetadata, Worker, Job, JobFault)
    for model in models:
        model.metadata.drop_all(engine)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from qonos.api import api
from qonos.common import exception
from qonos.common import timeutils
from qonos.common import utils


CONF = api.CONF


def serialize_metadata(metadata):
    to_return = {}
    for item in metadata:
        to_return[item['key']] = item['value']
    return to_return


def deserialize_metadata(metadata):
    return [{'key': key, 'value': value}
            for key, value in metadata.iteritems()]


def check_read_only_properties(values):
    _read_only_properties = ['created_at', 'updated_at']
    for key in values.keys():
        if key in _read_only_properties:
            msg = "%s is a read only attribute" % key
            raise exception.Forbidden(explanation=unicode(msg))

    return values


def serialize_schedule_metadata(schedule):
    metadata = schedule.pop('schedule_metadata')
    schedule['metadata'] = serialize_metadata(metadata)


def deserialize_schedule_metadata(schedule):
    if 'metadata' in schedule:
        metadata = schedule.pop('metadata')
        schedule['schedule_metadata'] = deserialize_metadata(metadata)


def serialize_job_metadata(job):
    metadata = job.pop('job_metadata')
    job['metadata'] = serialize_metadata(metadata)


def deserialize_job_metadata(job):
    if 'metadata' in job:
        metadata = job.pop('metadata')
        job['job_metadata'] = deserialize_metadata(metadata)


def schedule_to_next_run(schedule, start_time=None):
    start_time = start_time or timeutils.utcnow()
    minute = schedule.get('minute', '*')
    hour = schedule.get('hour', '*')
    day_of_month = schedule.get('day_of_month', '*')
    month = schedule.get('month', '*')
    day_of_week = schedule.get('day_of_week', '*')
    return utils.cron_string_to_next_datetime(minute, hour, day_of_month,
                                              month, day_of_week,
                                              start_time)


def get_new_timeout_by_action(action):
    now = timeutils.utcnow()

    group = 'action_' + action
    if group not in CONF:
        group = 'action_default'
    job_timeout_seconds = CONF.get(group).timeout_seconds
    return now + datetime.timedelta(seconds=job_timeout_seconds)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mox
import time

from qonos.common import timeutils
from qonos.qonosclient import exception as client_exc
from qonos.scheduler import scheduler
from qonos.tests.unit import utils as unit_utils
from qonos.tests import utils as test_utils


class TestScheduler(test_utils.BaseTestCase):

    def setUp(self):
        super(TestScheduler, self).setUp()
        self.mox = mox.Mox()
        self.client = self.mox.CreateMockAnything()

        def client_factory(*args, **kwargs):
            return self.client

        self.scheduler = scheduler.Scheduler(client_factory)

    def tearDown(self):
        super(TestScheduler, self).tearDown()

    def test_run_loop(self):
        self.config(job_schedule_interval=5, group='scheduler')
        timeutils.set_time_override()
        current_time = timeutils.isotime()
        called = {'enqueue_jobs': False}

        def fake(end_time=None):  # assert only end_time kwarg is passed
            self.assertEqual(end_time, current_time)
            called['enqueue_jobs'] = True

        self.stubs.Set(self.scheduler, 'enqueue_jobs', fake)
        fake_sleep = lambda x: None
        self.stubs.Set(time, 'sleep', fake_sleep)

        self.scheduler.run(run_once=True)
        self.assertTrue(called['enqueue_jobs'])

    def test_run_loop_take_too_long(self):
        self.config(job_schedule_interval=-1, group='scheduler')
        called = {'enqueue_jobs': False,
                  'log_warn': False}

        def fake(*args, **kwargs):
            called['enqueue_jobs'] = True

        self.stubs.Set(self.scheduler, 'enqueue_jobs', fake)
        fake_sleep = lambda x: None
        self.stubs.Set(time, 'sleep', fake_sleep)

        def fake_warn(msg, *args, **kwargs):
            called['log_warn'] = True

        self.stubs.Set(scheduler.LOG, 'warn', fake_warn)

        self.scheduler.run(run_once=True)
        self.assertTrue(called['enqueue_jobs'])
        self.assertTrue(called['log_warn'])

    def test_enqueue_jobs(self):
        called = {'get_schedules': False}
        next_run = '2010-11-30T17:00:00Z'

        def fake(*args, **kwargs):
            called['get_schedules'] = True
            return [{'id': unit_utils.SCHEDULE_UUID1, 'next_run': next_run}]

        self.stubs.Set(self.scheduler, 'get_schedules', fake)
        self.client.create_job(unit_utils.SCHEDULE_UUID1, next_run)
        self.mox.ReplayAll()
        self.scheduler.enqueue_jobs()
        self.mox.VerifyAll()

    def test_enqueue_jobs_job_exists(self):
        called = {'get_schedules': False}
        next_run = '2010-11-30T17:00:00Z'

        def fake(*args, **kwargs):
            called['get_schedules'] = True
            return [{'id': unit_utils.SCHEDULE_UUID1, 'next_run': next_run}]

        self.stubs.Set(self.scheduler, 'get_schedules', fake)
        self.client.create_job(unit_utils.SCHEDULE_UUID1,
                               next_run).AndRaise(
                                        client_exc.Duplicate())
        self.mox.ReplayAll()
        self.scheduler.enqueue_jobs()
        self.mox.VerifyAll()

    def test_get_schedules(self):
        timeutils.set_time_override()
        start_time = timeutils.isotime()
        timeutils.advance_time_seconds(30)
        end_time = timeutils.isotime()

        filter_args = {'next_run_after': start_time,
                       'next_run_before': end_time}
        self.client.list_schedules(filter_args=filter_args).AndReturn([])
        self.mox.ReplayAll()
        self.scheduler.get_schedules(start_time, end_time)
        self.mox.VerifyAll()

    def test_get_schedules_no_previous_run(self):
        end_time = timeutils.isotime()

        filter_args = {'next_run_before': end_time}
        self.client.list_schedules(filter_args=filter_args).AndReturn([])
        self.mox.ReplayAll()
        self.scheduler.get_schedules(end_time=end_time)
        self.mox.VerifyAll()



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import uuid
import webob.exc

from oslo.config import cfg

from qonos.api.v1 import workers
from qonos.common import exception
from qonos.common import timeutils
import qonos.db.simple.api as db_api
from qonos.tests.unit import utils as unit_utils
from qonos.tests import utils as test_utils

WORKER_ATTRS = ['id', 'host']


CONF = cfg.CONF


class TestWorkersApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestWorkersApi, self).setUp()
        self.controller = workers.WorkersController(db_api=db_api)
        self._create_workers()
        self._create_schedules()
        self._create_jobs()

    def tearDown(self):
        super(TestWorkersApi, self).tearDown()
        db_api.reset()

    def _create_workers(self):
        fixture = {'host': 'foo.cow', 'id': unit_utils.WORKER_UUID1}
        self.worker_1 = db_api.worker_create(fixture)
        fixture = {'host': 'bar.bar', 'id': unit_utils.WORKER_UUID2}
        self.worker_2 = db_api.worker_create(fixture)
        fixture = {'host': 'baz.cow', 'id': unit_utils.WORKER_UUID3}
        self.worker_3 = db_api.worker_create(fixture)
        fixture = {'host': 'qux.bar', 'id': unit_utils.WORKER_UUID4}
        self.worker_4 = db_api.worker_create(fixture)

    def _create_schedules(self):
        fixture = {
            'id': unit_utils.SCHEDULE_UUID1,
            'tenant': unit_utils.TENANT1,
            'action': 'snapshot',
            'minute': '30',
            'hour': '2',
            'next_run': '2012-11-27T02:30:00Z'
        }
        self.schedule_1 = db_api.schedule_create(fixture)
        fixture = {
            'id': unit_utils.SCHEDULE_UUID2,
            'tenant': unit_utils.TENANT2,
            'action': 'snapshot',
            'minute': '30',
            'hour': '2',
            'next_run': '2012-11-27T02:30:00Z',
            'schedule_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        self.schedule_2 = db_api.schedule_create(fixture)
        fixture = {
            'id': unit_utils.SCHEDULE_UUID3,
            'tenant': unit_utils.TENANT3,
            'action': 'snapshot',
            'minute': '30',
            'hour': '4',
            'next_run': '2012-11-27T02:30:00Z',
        }
        self.schedule_3 = db_api.schedule_create(fixture)
        fixture = {
            'id': unit_utils.SCHEDULE_UUID4,
            'tenant': unit_utils.TENANT4,
            'action': 'snapshot',
            'minute': '30',
            'hour': '5',
            'next_run': '2012-11-27T02:30:00Z',
        }
        self.schedule_4 = db_api.schedule_create(fixture)

    def _create_jobs(self):
        now = timeutils.utcnow()
        timeout = now + datetime.timedelta(hours=1)
        hard_timeout = now + datetime.timedelta(hours=4)

        fixture = {
            'id': unit_utils.JOB_UUID1,
            'schedule_id': self.schedule_1['id'],
            'tenant': unit_utils.TENANT1,
            'worker_id': None,
            'action': 'snapshot',
            'status': None,
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 0,
        }
        self.job_1 = db_api.job_create(fixture)
        fixture = {
            'id': unit_utils.JOB_UUID2,
            'schedule_id': self.schedule_2['id'],
            'tenant': unit_utils.TENANT2,
            'worker_id': unit_utils.WORKER_UUID2,
            'action': 'snapshot',
            'status': None,
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 1,
            'job_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        self.job_2 = db_api.job_create(fixture)
        fixture = {
            'id': unit_utils.JOB_UUID3,
            'schedule_id': self.schedule_3['id'],
            'tenant': unit_utils.TENANT3,
            'worker_id': unit_utils.WORKER_UUID2,
            'action': 'snapshot',
            'status': None,
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 1,
            'job_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        self.job_3 = db_api.job_create(fixture)
        fixture = {
            'id': unit_utils.JOB_UUID4,
            'schedule_id': self.schedule_4['id'],
            'tenant': unit_utils.TENANT4,
            'worker_id': unit_utils.WORKER_UUID2,
            'action': 'snapshot',
            'status': None,
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 1,
            'job_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        self.job_4 = db_api.job_create(fixture)

    def test_list(self):
        self.config(limit_param_default=2, api_limit_max=4)
        request = unit_utils.get_fake_request(method='GET')
        workers = self.controller.list(request).get('workers')
        self.assertEqual(len(workers), 2)
        for k in WORKER_ATTRS:
            self.assertEqual(set([s[k] for s in workers]),
                             set([self.worker_1[k], self.worker_2[k]]))

    def test_list_limit(self):
        path = '?limit=2'
        request = unit_utils.get_fake_request(path=path, method='GET')
        workers = self.controller.list(request).get('workers')
        self.assertEqual(len(workers), 2)

    def test_list_limit_max(self):
        self.config(api_limit_max=3)
        path = '?limit=4'
        request = unit_utils.get_fake_request(path=path, method='GET')
        workers = self.controller.list(request).get('workers')
        self.assertEqual(len(workers), 3)

    def test_list_default_limit(self):
        self.config(limit_param_default=2)
        request = unit_utils.get_fake_request(method='GET')
        workers = self.controller.list(request).get('workers')
        self.assertEqual(len(workers), 2)

    def test_list_with_marker(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % unit_utils.WORKER_UUID1
        request = unit_utils.get_fake_request(path=path, method='GET')
        workers = self.controller.list(request).get('workers')
        self.assertEqual(len(workers), 2)
        for k in WORKER_ATTRS:
            self.assertEqual(set([s[k] for s in workers]),
                             set([self.worker_2[k], self.worker_3[k]]))

    def test_list_marker_not_specified(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % ''
        request = unit_utils.get_fake_request(path=path, method='GET')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.list, request)

    def test_list_marker_not_found(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % '3c5817e2-76cb-41fe-b012-2935e406db87'
        request = unit_utils.get_fake_request(path=path, method='GET')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.list, request)

    def test_list_invalid_marker(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % '3c5817e2-76cb'
        request = unit_utils.get_fake_request(path=path, method='GET')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.list, request)

    def test_list_with_limit_and_marker(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s&limit=1' % unit_utils.WORKER_UUID1
        request = unit_utils.get_fake_request(path=path, method='GET')
        workers = self.controller.list(request).get('workers')
        self.assertEqual(len(workers), 1)
        for k in WORKER_ATTRS:
            self.assertEqual(set([s[k] for s in workers]),
                             set([self.worker_2[k]]))

    def test_get(self):
        request = unit_utils.get_fake_request(method='GET')
        actual = self.controller.get(request,
                                     self.worker_1['id']).get('worker')
        for k in WORKER_ATTRS:
            self.assertEqual(actual[k], self.worker_1[k])

    def test_get_not_found(self):
        request = unit_utils.get_fake_request(method='GET')
        worker_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.get, request, worker_id)

    def test_create(self):
        request = unit_utils.get_fake_request(method='POST')
        host = 'blah'
        fixture = {'worker': {'host': host, 'id': unit_utils.WORKER_UUID5}}
        actual = self.controller.create(request, fixture)['worker']
        self.assertEqual(host, actual['host'])

    def test_delete(self):
        request = unit_utils.get_fake_request(method='GET')
        request = unit_utils.get_fake_request(method='DELETE')
        self.controller.delete(request, self.worker_1['id'])
        self.assertRaises(exception.NotFound, db_api.worker_get_by_id,
                          self.worker_1['id'])

    def test_delete_not_found(self):
        request = unit_utils.get_fake_request(method='DELETE')
        worker_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.delete, request, worker_id)

    def test_get_next_job_unknown_worker_for_action(self):
        request = unit_utils.get_fake_request(method='POST')
        fixture = {'action': 'snapshot'}
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.get_next_job,
                          request, 'unknown_worker_id', fixture)

    def test_get_next_job_none_for_action(self):
        request = unit_utils.get_fake_request(method='POST')
        fixture = {'action': 'dummy'}
        job = self.controller.get_next_job(request,
                                           self.worker_1['id'],
                                           fixture)
        self.assertEqual(job['job'], None)

    def test_get_next_job_for_action(self):
        request = unit_utils.get_fake_request(method='POST')
        fixture = {'action': 'snapshot'}
        job = self.controller.get_next_job(request,
                                           self.worker_1['id'],
                                           fixture)
        self.assertEqual(self.worker_1['id'], job['job']['worker_id'])



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo.config import cfg

from qonos.openstack.common import importutils

sql_connection_opt = cfg.StrOpt('sql_connection',
                                default='sqlite:///qonos.sqlite',
                                secret=True,
                                metavar='CONNECTION',
                                help='A valid SQLAlchemy connection '
                                     'string for the database. '
                                     'Default: %default')

CONF = cfg.CONF
CONF.register_opt(sql_connection_opt)


def get_api():
    db_api = importutils.import_module(CONF.db_api)
    db_api.configure_db()
    return db_api



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
import operator
import uuid

from operator import itemgetter
from qonos.common import exception
from qonos.common import timeutils
import qonos.db.db_utils as db_utils
from qonos.openstack.common.gettextutils import _


DATA = {
    'schedules': {},
    'schedule_metadata': {},
    'jobs': {},
    'job_metadata': {},
    'workers': {},
    'job_faults': {},
}


def configure_db():
    pass


def reset():
    global DATA
    for k in DATA:
        DATA[k] = {}


def _gen_base_attributes(item_id=None):
    values = {}
    if item_id is None:
        values['id'] = str(uuid.uuid4())
    values['created_at'] = timeutils.utcnow()
    values['updated_at'] = timeutils.utcnow()
    return copy.deepcopy(values)


def _schedule_create(values):
    global DATA
    DATA['schedules'][values['id']] = values
    _schedule_meta_init(values['id'])
    return copy.deepcopy(values)


def _do_pagination(items, marker, limit):
    """
    This method mimics the behavior of sqlalchemy paginate_query.
    It takes items and pagination parameters - 'limit' and 'marker'
    to filter out the items to be returned. Items are sorted in
    lexicographical order based on the sort key - 'id'.
    """
    items = sorted(items, key=itemgetter('id'))
    start = 0
    end = -1
    if marker is None:
        start = 0
    else:
        for i, item in enumerate(items):
            if item['id'] == marker:
                start = i + 1
                break
        else:
            msg = _('Marker %s not found') % marker
            raise exception.NotFound(explanation=msg)

    end = start + limit if limit is not None else None
    return items[start:end]


def schedule_get_all(filter_args={}):
    SCHEDULE_BASE_FILTERS = ['next_run_after', 'next_run_before',
                             'tenant', 'limit', 'marker']
    schedules = copy.deepcopy(DATA['schedules'].values())
    schedules_mutate = copy.deepcopy(DATA['schedules'].values())

    for schedule in schedules:
        schedule['schedule_metadata'] =\
            copy.deepcopy(schedule_meta_get_all(schedule['id']))

    schedules_mutate = copy.deepcopy(schedules)

    if 'next_run_before' in filter_args:
        for schedule in schedules:
            if not schedule['next_run'] <= filter_args['next_run_before']:
                if schedule in schedules_mutate:
                    del schedules_mutate[schedules_mutate.index(schedule)]
        filter_args.pop('next_run_before')

    if 'next_run_after' in filter_args:
        for schedule in schedules:
            if not schedule['next_run'] >= filter_args['next_run_after']:
                if schedule in schedules_mutate:
                    del schedules_mutate[schedules_mutate.index(schedule)]
        filter_args.pop('next_run_after')

    if filter_args.get('tenant') is not None:
        for schedule in schedules:
            if schedule['tenant'] != filter_args['tenant']:
                if schedule in schedules_mutate:
                    del schedules_mutate[schedules_mutate.index(schedule)]
        filter_args.pop('tenant')

    if filter_args.get('action') is not None:
        for schedule in schedules:
            if schedule['action'] != filter_args['action']:
                if schedule in schedules_mutate:
                    del schedules_mutate[schedules_mutate.index(schedule)]
        filter_args.pop('action')

    for filter_key in filter_args.keys():
        if filter_key not in SCHEDULE_BASE_FILTERS:
            filter_value = filter_args[filter_key]
            for schedule in schedules:
                if schedule['schedule_metadata']:
                    for schedule_metadata in schedule['schedule_metadata']:
                        if not(schedule_metadata['key'] == filter_key and
                               schedule_metadata['value'] == filter_value):
                            try:
                                schedule_mutated = (
                                        schedules_mutate.index(schedule))
                                del schedules_mutate[schedule_mutated]
                            except Exception:
                                pass
                            break
                else:
                    if schedule in schedules_mutate:
                        del schedules_mutate[schedules_mutate.index(schedule)]

    marker = filter_args.get('marker')
    limit = filter_args.get('limit')
    schedules_mutate = _do_pagination(schedules_mutate, marker, limit)
    return schedules_mutate


def schedule_get_by_id(schedule_id):
    if schedule_id not in DATA['schedules']:
        raise exception.NotFound()
    schedule = copy.deepcopy(DATA['schedules'][schedule_id])
    schedule['schedule_metadata'] = \
        copy.deepcopy(schedule_meta_get_all(schedule_id))
    return schedule


def schedule_create(schedule_values):
    db_utils.validate_schedule_values(schedule_values)
    values = copy.deepcopy(schedule_values)
    schedule = {}

    metadata = []
    if 'schedule_metadata' in values:
        metadata = values['schedule_metadata']
        del values['schedule_metadata']

    schedule.update(values)
    item_id = values.get('id')
    schedule.update(_gen_base_attributes(item_id=item_id))
    schedule = _schedule_create(schedule)

    for metadatum in metadata:
        schedule_meta_create(schedule['id'], metadatum)

    return schedule_get_by_id(schedule['id'])


def schedule_update(schedule_id, schedule_values):
    global DATA
    values = schedule_values.copy()

    if schedule_id not in DATA['schedules']:
        raise exception.NotFound()

    metadata = None
    if 'schedule_metadata' in values:
        metadata = values['schedule_metadata']
        del values['schedule_metadata']

    if len(values) > 0:
        schedule = DATA['schedules'][schedule_id]
        schedule['updated_at'] = timeutils.utcnow()
        schedule.update(values)

    if metadata is not None:
        DATA['schedule_metadata'][schedule_id] = {}
        for metadatum in metadata:
            schedule_meta_create(schedule_id, metadatum)

    return schedule_get_by_id(schedule_id)


def schedule_test_and_set_next_run(schedule_id, expected_next_run, next_run):
    global DATA

    schedule = DATA['schedules'].get(schedule_id)

    if not schedule:
        raise exception.NotFound()

    if expected_next_run:
        expected_next_run = expected_next_run.replace(tzinfo=None)
        current_next_run = schedule.get('next_run')
        if current_next_run:
            current_next_run = current_next_run.replace(tzinfo=None)
        if expected_next_run != current_next_run:
            raise exception.NotFound()

    schedule['next_run'] = next_run


def schedule_delete(schedule_id):
    global DATA
    if schedule_id not in DATA['schedules']:
        raise exception.NotFound()
    del DATA['schedules'][schedule_id]


def _schedule_meta_init(schedule_id):
    if DATA['schedule_metadata'].get(schedule_id) is None:
        DATA['schedule_metadata'][schedule_id] = {}


def schedule_meta_create(schedule_id, values):
    global DATA
    if DATA['schedules'].get(schedule_id) is None:
        msg = _('Schedule %s could not be found') % schedule_id
        raise exception.NotFound(message=msg)

    _schedule_meta_init(schedule_id)

    try:
        _check_meta_exists(schedule_id, values['key'])
    except exception.NotFound:
        pass
    else:
        raise exception.Duplicate()

    meta = {}
    values['schedule_id'] = schedule_id
    meta.update(values)
    meta.update(_gen_base_attributes())
    DATA['schedule_metadata'][schedule_id][values['key']] = meta
    return copy.deepcopy(meta)


def _check_schedule_exists(schedule_id):
    if DATA['schedules'].get(schedule_id) is None:
        msg = _('Schedule %s could not be found') % schedule_id
        raise exception.NotFound(message=msg)


def _check_meta_exists(schedule_id, key):
    _check_schedule_exists(schedule_id)

    if (DATA['schedule_metadata'].get(schedule_id) is None or
            DATA['schedule_metadata'][schedule_id].get(key) is None):
        msg = (_('Meta %(key)s could not be found '
                 'for Schedule %(schedule_id)s ') %
               {'key': key, 'schedule_id': schedule_id})
        raise exception.NotFound(message=msg)


def schedule_meta_get_all(schedule_id):
    _check_schedule_exists(schedule_id)
    _schedule_meta_init(schedule_id)

    return DATA['schedule_metadata'][schedule_id].values()


def schedule_metadata_update(schedule_id, values):
    global DATA
    if DATA['schedules'].get(schedule_id) is None:
        msg = _('Schedule %s could not be found') % schedule_id
        raise exception.NotFound(message=msg)

    DATA['schedule_metadata'][schedule_id] = {}
    for metadatum in values:
        schedule_meta_create(schedule_id, metadatum)

    return copy.deepcopy(DATA['schedule_metadata'][schedule_id].values())


def _delete_schedule_meta(schedule_id, key):
    del DATA['schedule_metadata'][schedule_id][key]


def schedule_meta_delete(schedule_id, key):
    _check_meta_exists(schedule_id, key)
    _delete_schedule_meta(schedule_id, key)


def worker_get_all(params={}):
    workers = copy.deepcopy(DATA['workers'].values())
    marker = params.get('marker')
    limit = params.get('limit')
    workers = _do_pagination(workers, marker, limit)
    return workers


def worker_get_by_id(worker_id):
    if worker_id not in DATA['workers']:
        raise exception.NotFound()
    return copy.deepcopy(DATA['workers'][worker_id])


def worker_create(values):
    global DATA
    worker_values = copy.deepcopy(values)
    if 'process_id' not in worker_values:
        worker_values['process_id'] = None
    worker = {}
    worker.update(worker_values)
    item_id = values.get('id')
    worker.update(_gen_base_attributes(item_id=item_id))
    DATA['workers'][worker['id']] = worker
    return copy.deepcopy(worker)


def worker_delete(worker_id):
    global DATA
    if worker_id not in DATA['workers']:
        raise exception.NotFound()
    del DATA['workers'][worker_id]


def job_create(job_values):
    global DATA
    db_utils.validate_job_values(job_values)
    values = job_values.copy()
    job = {}

    metadata = []
    if 'job_metadata' in values:
        metadata = values['job_metadata']
        del values['job_metadata']

    if not 'retry_count' in values:
        values['retry_count'] = 0
    job['worker_id'] = None

    job.update(values)
    item_id = values.get('id')
    job.update(_gen_base_attributes(item_id=item_id))

    DATA['jobs'][job['id']] = job

    for metadatum in metadata:
        job_meta_create(job['id'], metadatum)

    return job_get_by_id(job['id'])


def job_get_all(params={}):
    jobs = copy.deepcopy(DATA['jobs'].values())

    for job in jobs:
        job['job_metadata'] =\
            job_meta_get_all_by_job_id(job['id'])

    marker = params.get('marker')
    limit = params.get('limit')
    jobs = _do_pagination(jobs, marker, limit)

    return jobs


def job_get_by_id(job_id):
    if job_id not in DATA['jobs']:
        raise exception.NotFound()

    job = copy.deepcopy(DATA['jobs'][job_id])

    job['job_metadata'] = \
        job_meta_get_all_by_job_id(job_id)

    return job


def job_updated_at_get_by_id(job_id):
    job = job_get_by_id(job_id)
    return job['updated_at']


def job_get_and_assign_next_by_action(action, worker_id, max_retry,
                                      new_timeout):
    """Get the next available job for the given action and assign it
    to the worker for worker_id.
    This must be an atomic action!"""
    job_ref = None
    now = timeutils.utcnow()
    jobs = _jobs_get_sorted()
    for job in jobs:
        if job['action'] == action and \
                job['retry_count'] < max_retry and \
                job['hard_timeout'] > now and \
                job['status'] not in ['DONE', 'CANCELLED'] and \
                (job['worker_id'] is None or job['timeout'] <= now):
            job_ref = job
            break

    if job_ref is None:
        return None

    job_id = job_ref['id']
    DATA['jobs'][job_id]['worker_id'] = worker_id
    DATA['jobs'][job_id]['timeout'] = new_timeout
    DATA['jobs'][job_id]['retry_count'] = job_ref['retry_count'] + 1
    job = copy.deepcopy(DATA['jobs'][job_id])
    job['job_metadata'] = job_meta_get_all_by_job_id(job_id)

    return job


def _jobs_get_sorted():
    jobs = copy.deepcopy(DATA['jobs'])
    sorted_jobs = []
    for job_id in jobs:
        sorted_jobs.append(jobs[job_id])

    sorted_jobs = sorted(sorted_jobs, key=operator.itemgetter('created_at'))
    return sorted_jobs


def _jobs_cleanup_hard_timed_out():
    """Find all jobs with hard_timeout values which have passed
    and delete them, logging the timeout / failure as appropriate"""
    now = timeutils.utcnow()
    del_ids = []
    for job_id in DATA['jobs']:
        job = DATA['jobs'][job_id]
        print now, job['hard_timeout']
        print now - job['hard_timeout']
        if (now - job['hard_timeout']) > datetime.timedelta(microseconds=0):
            del_ids.append(job_id)

    for job_id in del_ids:
        job_delete(job_id)
    return len(del_ids)


def job_update(job_id, job_values):
    global DATA
    values = job_values.copy()
    if job_id not in DATA['jobs']:
        raise exception.NotFound()

    metadata = None
    if 'job_metadata' in values:
        metadata = values['job_metadata']
        del values['job_metadata']

    if len(values) > 0:
        job = DATA['jobs'][job_id]
        #NOTE(ameade): This must come before update specified values since
        # we may be trying to manually set updated_at
        job['updated_at'] = timeutils.utcnow()
        job.update(values)

    if metadata is not None:
        DATA['job_metadata'][job_id] = {}
        for metadatum in metadata:
            job_meta_create(job_id, metadatum)

    return job_get_by_id(job_id)


def job_delete(job_id):
    global DATA
    if job_id not in DATA['jobs']:
        raise exception.NotFound()
    del DATA['jobs'][job_id]


def job_meta_create(job_id, values):
    global DATA
    values['job_id'] = job_id
    _check_job_exists(job_id)

    if DATA['job_metadata'].get(job_id) is None:
        DATA['job_metadata'][job_id] = {}

    try:
        _check_job_meta_exists(job_id, values['key'])
    except exception.NotFound:
        pass
    else:
        raise exception.Duplicate()

    meta = {}
    meta.update(values)
    meta.update(_gen_base_attributes())
    DATA['job_metadata'][job_id][values['key']] = meta
    return copy.deepcopy(meta)


def _check_job_exists(job_id):
    if not job_id in DATA['jobs']:
        msg = _('Job %s could not be found') % job_id
        raise exception.NotFound(message=msg)


def _check_job_meta_exists(job_id, key):

    if (DATA['job_metadata'].get(job_id) is None or
            DATA['job_metadata'][job_id].get(key) is None):
        msg = (_('Meta %(key)s could not be found for Job %(job_id)s ') %
               {'key': key, 'job_id': job_id})
        raise exception.NotFound(message=msg)


def job_meta_get_all_by_job_id(job_id):
    _check_job_exists(job_id)

    if job_id not in DATA['job_metadata']:
        DATA['job_metadata'][job_id] = {}

    return copy.deepcopy(DATA['job_metadata'][job_id].values())


def job_metadata_update(job_id, values):
    global DATA
    _check_job_exists(job_id)

    DATA['job_metadata'][job_id] = {}
    for metadatum in values:
        job_meta_create(job_id, metadatum)

    return copy.deepcopy(DATA['job_metadata'][job_id].values())


def _job_faults_get_sorted():
    jobs = copy.deepcopy(DATA['job_faults'])
    sorted_jobs = []
    for job_id in jobs:
        sorted_jobs.append(jobs[job_id])

    sorted_jobs = sorted(sorted_jobs, key=operator.itemgetter('created_at'),
                         reverse=True)
    return sorted_jobs


def job_fault_latest_for_job_id(job_id):
    job_faults = _job_faults_get_sorted()
    for job_fault in job_faults:
        if job_fault['job_id'] == job_id:
            return job_fault
    return None


def job_fault_create(values):
    global DATA
    job_fault = {}
    job_fault.update(values)
    item_id = values.get('id')
    job_fault.update(_gen_base_attributes(item_id=item_id))
    DATA['job_faults'][job_fault['id']] = job_fault
    return copy.deepcopy(job_fault)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import uuid
import webob.exc

from oslo.config import cfg

from qonos.api.v1 import api_utils
from qonos.api.v1 import schedules
from qonos.common import exception
from qonos.common import timeutils
from qonos.common import utils as qonos_utils
from qonos.db.simple import api as db_api
from qonos.tests.unit import utils as unit_utils
from qonos.tests import utils as test_utils


SCHEDULE_ATTRS = ['id', 'tenant', 'action',
                  'minute', 'hour']


CONF = cfg.CONF


class TestSchedulesApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestSchedulesApi, self).setUp()
        timeutils.set_time_override(datetime.datetime(2013, 2, 17, 2, 0, 0, 0))
        self.controller = schedules.SchedulesController(db_api=db_api)
        self._create_schedules()

    def tearDown(self):
        super(TestSchedulesApi, self).tearDown()
        timeutils.clear_time_override()
        db_api.reset()

    def _create_schedules(self):
        fixture = {
            'id': unit_utils.SCHEDULE_UUID1,
            'tenant': unit_utils.TENANT1,
            'action': 'snapshot',
            'minute': '30',
            'hour': '2',
            'next_run': qonos_utils.cron_string_to_next_datetime(30, 2)
        }
        self.schedule_1 = db_api.schedule_create(fixture)
        fixture = {
            'id': unit_utils.SCHEDULE_UUID2,
            'tenant': unit_utils.TENANT2,
            'action': 'snapshot',
            'minute': '30',
            'hour': '3',
            'next_run': qonos_utils.cron_string_to_next_datetime(30, 3)
        }
        self.schedule_2 = db_api.schedule_create(fixture)
        fixture = {
            'id': unit_utils.SCHEDULE_UUID3,
            'tenant': unit_utils.TENANT3,
            'action': 'snapshot',
            'minute': '30',
            'hour': '4',
            'next_run': qonos_utils.cron_string_to_next_datetime(30, 4)
        }
        self.schedule_3 = db_api.schedule_create(fixture)
        fixture = {
            'id': unit_utils.SCHEDULE_UUID4,
            'tenant': unit_utils.TENANT4,
            'action': 'other',
            'minute': '30',
            'hour': '5',
            'next_run': qonos_utils.cron_string_to_next_datetime(30, 5)
        }
        self.schedule_4 = db_api.schedule_create(fixture)

    def test_list(self):
        request = unit_utils.get_fake_request(method='GET')
        schedules = self.controller.list(request).get('schedules')
        links = self.controller.list(request).get('schedules_links')
        self.assertEqual(len(schedules), 4)
        for k in SCHEDULE_ATTRS:
            self.assertEqual(set([s[k] for s in schedules]),
                             set([self.schedule_1[k], self.schedule_2[k],
                                  self.schedule_3[k], self.schedule_4[k]]))
        for item in links:
            if item.get('rel') == 'next':
                self.assertEqual(item.get('href'), None)

    def test_list_next_run_filtered(self):
        next_run = self.schedule_1['next_run']
        path = '?next_run_after=%s&next_run_before=%s'
        path = path % (next_run, next_run)
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 1)

    def test_list_next_run_after_filtered(self):
        next_run = self.schedule_1['next_run']
        path = '?next_run_after=%s'
        path = path % next_run
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 4)

    def test_list_next_run_before_filtered(self):
        next_run = self.schedule_3['next_run']
        path = '?next_run_before=%s'
        path = path % next_run
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 3)

    def test_list_next_run_filtered_before_less_than_after(self):
        after = self.schedule_3['next_run']
        before = timeutils.isotime(after
                        - datetime.timedelta(seconds=1))
        path = '?next_run_after=%s&next_run_before=%s'
        path = path % (after, before)
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 0)

    def test_list_limit(self):
        path = '?limit=2'
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 2)

    def test_list_filter_by_action(self):
        path = '?action=snapshot'
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 3)

    def test_list_limit_max(self):
        self.config(api_limit_max=3)
        path = '?limit=4'
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 3)

    def test_list_default_limit(self):
        self.config(limit_param_default=2)
        request = unit_utils.get_fake_request(method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 2)

    def test_list_with_marker(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % unit_utils.SCHEDULE_UUID1
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 2)
        for k in SCHEDULE_ATTRS:
            self.assertEqual(set([s[k] for s in schedules]),
                             set([self.schedule_2[k], self.schedule_3[k]]))

    def test_list_marker_not_specified(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % ''
        request = unit_utils.get_fake_request(path=path, method='GET')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.list, request)

    def test_list_marker_not_found(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % '3c5817e2-76cb-41fe-b012-2935e406db87'
        request = unit_utils.get_fake_request(path=path, method='GET')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.list, request)

    def test_list_invalid_marker(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % '3c5817e2-76cb'
        request = unit_utils.get_fake_request(path=path, method='GET')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.list, request)

    def test_list_with_limit_and_marker(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s&limit=1' % unit_utils.SCHEDULE_UUID1
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        self.assertEqual(len(schedules), 1)
        for k in SCHEDULE_ATTRS:
            self.assertEqual(set([s[k] for s in schedules]),
                             set([self.schedule_2[k]]))

    def test_list_schedules_links(self):
        self.config(limit_param_default=2, api_limit_max=4)
        path = '?marker=%s' % unit_utils.SCHEDULE_UUID1
        request = unit_utils.get_fake_request(path=path, method='GET')
        schedules = self.controller.list(request).get('schedules')
        links = self.controller.list(request).get('schedules_links')
        self.assertEqual(len(schedules), 2)
        for k in SCHEDULE_ATTRS:
            self.assertEqual(set([s[k] for s in schedules]),
                             set([self.schedule_2[k], self.schedule_3[k]]))
        for item in links:
            if item.get('rel') == 'next':
                marker = unit_utils.SCHEDULE_UUID3
                self.assertEqual(item.get('href'), '/v1/schedules?marker=%s' %
                                 marker)

    def test_get(self):
        request = unit_utils.get_fake_request(method='GET')
        actual = self.controller.get(request,
                                     self.schedule_1['id']).get('schedule')
        for k in SCHEDULE_ATTRS:
            self.assertEqual(actual[k], self.schedule_1[k])

    def test_get_not_found(self):
        request = unit_utils.get_fake_request(method='GET')
        schedule_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.get, request, schedule_id)

    def test_create(self):
        fixture = {'schedule': {
            'id': unit_utils.SCHEDULE_UUID5,
            'tenant': unit_utils.TENANT1,
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
        }}
        expected = fixture['schedule']
        request = unit_utils.get_fake_request(method='POST')

        actual = self.controller.create(request, fixture)['schedule']

        self.assertNotEqual(actual.get('id'), None)
        self.assertNotEqual(actual.get('created_at'), None)
        self.assertNotEqual(actual.get('updated_at'), None)
        self.assertNotEqual(actual.get('next_run'), None)
        self.assertEqual(expected['tenant'], actual['tenant'])
        self.assertEqual(expected['action'], actual['action'])
        self.assertEqual(expected['minute'], actual['minute'])
        self.assertEqual(expected['hour'], actual['hour'])

    def test_create_zero_hour(self):
        hour = 0
        fixture = {'schedule': {
            'id': unit_utils.SCHEDULE_UUID5,
            'tenant': unit_utils.TENANT1,
            'action': 'snapshot',
            'minute': 30,
            'hour': hour,
        }}
        expected = fixture['schedule']
        request = unit_utils.get_fake_request(method='POST')

        actual = self.controller.create(request, fixture)['schedule']

        self.assertNotEqual(actual.get('id'), None)
        self.assertNotEqual(actual.get('created_at'), None)
        self.assertNotEqual(actual.get('updated_at'), None)
        now = timeutils.utcnow()
        if not (now.hour == hour and now.minute < 30):
            now = now + datetime.timedelta(days=1)
        expected_next_run = timeutils.isotime(
            now.replace(hour=hour, minute=30, second=0,
                        microsecond=0))
        self.assertEqual(expected_next_run, actual['next_run'])
        self.assertEqual(expected['tenant'], actual['tenant'])
        self.assertEqual(expected['action'], actual['action'])
        self.assertEqual(expected['minute'], actual['minute'])
        self.assertEqual(expected['hour'], actual['hour'])

    def test_create_no_body_bad_request(self):
        request = unit_utils.get_fake_request(method='POST')
        schedule_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          request, None)

    def test_create_malformed_body_bad_request(self):
        request = unit_utils.get_fake_request(method='POST')
        schedule_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          request, 'fake-body')

    def test_create_no_schedule_bad_request(self):
        request = unit_utils.get_fake_request(method='POST')
        schedule_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.create,
                          request, {'minute': '5'})

    def test_delete(self):
        request = unit_utils.get_fake_request(method='GET')
        request = unit_utils.get_fake_request(method='DELETE')
        self.controller.delete(request, self.schedule_1['id'])
        self.assertRaises(exception.NotFound, db_api.worker_get_by_id,
                          self.schedule_1['id'])

    def test_delete_not_found(self):
        request = unit_utils.get_fake_request(method='DELETE')
        schedule_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.delete, request, schedule_id)

    def test_update(self):

        expected_next_run = '1989-01-19T12:00:00Z'

        def fake_schedule_to_next_run(*args, **kwargs):
            return timeutils.parse_isotime(expected_next_run)

        self.stubs.Set(api_utils, 'schedule_to_next_run',
                       fake_schedule_to_next_run)

        request = unit_utils.get_fake_request(method='PUT')
        update_fixture = {'schedule': {
                            'minute': '55',
                            'hour': '5',
                            'day': '2',
                            'day_of_week': '4',
                            'day_of_month': '23',
                            'action': 'new-action',
                            'tenant': 'new-tenant',
                         }}

        updated = self.controller.update(request, self.schedule_1['id'],
                                         update_fixture)['schedule']

        self.assertEqual(update_fixture['schedule']['minute'],
                         updated['minute'])
        self.assertEqual(update_fixture['schedule']['hour'],
                         updated['hour'])
        self.assertEqual(update_fixture['schedule']['day'],
                         updated['day'])
        self.assertEqual(update_fixture['schedule']['day_of_week'],
                         updated['day_of_week'])
        self.assertEqual(update_fixture['schedule']['day_of_month'],
                         updated['day_of_month'])
        self.assertEqual(update_fixture['schedule']['action'],
                         updated['action'])
        self.assertEqual(update_fixture['schedule']['tenant'],
                         updated['tenant'])
        self.assertEqual(updated['next_run'], expected_next_run)

    def test_update_with_hour(self):

        expected_next_run = '1989-01-19T12:00:00Z'

        def fake_schedule_to_next_run(*args, **kwargs):
            return timeutils.parse_isotime(expected_next_run)

        self.stubs.Set(api_utils, 'schedule_to_next_run',
                       fake_schedule_to_next_run)

        request = unit_utils.get_fake_request(method='PUT')
        update_fixture = {'schedule': {'hour': '5'}}

        updated = self.controller.update(request, self.schedule_1['id'],
                                         update_fixture)['schedule']

        self.assertNotEqual(updated.get('created_at'), None)
        self.assertNotEqual(updated.get('updated_at'), None)
        self.assertEqual(self.schedule_1['tenant'], updated['tenant'])
        self.assertEqual(self.schedule_1['action'], updated['action'])
        self.assertEqual(update_fixture['schedule']['hour'],
                         updated['hour'])
        self.assertFalse(updated['minute'])
        self.assertFalse(updated['month'])
        self.assertFalse(updated['day_of_week'])
        self.assertFalse(updated['day_of_month'])
        self.assertEqual(updated['next_run'], expected_next_run)

    def test_update_with_minute(self):

        expected_next_run = '1989-01-19T12:00:00Z'

        def fake_schedule_to_next_run(*args, **kwargs):
            return timeutils.parse_isotime(expected_next_run)

        self.stubs.Set(api_utils, 'schedule_to_next_run',
                       fake_schedule_to_next_run)

        request = unit_utils.get_fake_request(method='PUT')
        update_fixture = {'schedule': {'minute': '5'}}

        updated = self.controller.update(request, self.schedule_1['id'],
                                         update_fixture)['schedule']

        self.assertNotEqual(updated.get('created_at'), None)
        self.assertNotEqual(updated.get('updated_at'), None)
        self.assertEqual(self.schedule_1['tenant'], updated['tenant'])
        self.assertEqual(self.schedule_1['action'], updated['action'])
        self.assertFalse(updated['hour'])
        self.assertFalse(updated['month'])
        self.assertFalse(updated['day_of_week'])
        self.assertFalse(updated['day_of_month'])
        self.assertEqual(update_fixture['schedule']['minute'],
                         updated['minute'])
        self.assertEqual(updated['next_run'], expected_next_run)

    def test_update_with_month(self):

        expected_next_run = '1989-01-19T12:00:00Z'

        def fake_schedule_to_next_run(*args, **kwargs):
            return timeutils.parse_isotime(expected_next_run)

        self.stubs.Set(api_utils, 'schedule_to_next_run',
                       fake_schedule_to_next_run)

        request = unit_utils.get_fake_request(method='PUT')
        update_fixture = {'schedule': {'month': '5'}}

        updated = self.controller.update(request, self.schedule_1['id'],
                                         update_fixture)['schedule']

        self.assertNotEqual(updated.get('created_at'), None)
        self.assertNotEqual(updated.get('updated_at'), None)
        self.assertEqual(self.schedule_1['tenant'], updated['tenant'])
        self.assertEqual(self.schedule_1['action'], updated['action'])
        self.assertEqual(update_fixture['schedule']['month'],
                         updated['month'])
        self.assertFalse(updated['minute'])
        self.assertFalse(updated['hour'])
        self.assertFalse(updated['day_of_week'])
        self.assertFalse(updated['day_of_month'])
        self.assertEqual(updated['next_run'], expected_next_run)

    def test_update_with_day_of_week(self):

        expected_next_run = '1989-01-19T12:00:00Z'

        def fake_schedule_to_next_run(*args, **kwargs):
            return timeutils.parse_isotime(expected_next_run)

        self.stubs.Set(api_utils, 'schedule_to_next_run',
                       fake_schedule_to_next_run)

        request = unit_utils.get_fake_request(method='PUT')
        update_fixture = {'schedule': {'day_of_week': '5'}}

        updated = self.controller.update(request, self.schedule_1['id'],
                                         update_fixture)['schedule']

        self.assertNotEqual(updated.get('created_at'), None)
        self.assertNotEqual(updated.get('updated_at'), None)
        self.assertEqual(self.schedule_1['tenant'], updated['tenant'])
        self.assertEqual(self.schedule_1['action'], updated['action'])
        self.assertEqual(update_fixture['schedule']['day_of_week'],
                         updated['day_of_week'])
        self.assertFalse(updated['minute'])
        self.assertFalse(updated['hour'])
        self.assertFalse(updated['month'])
        self.assertFalse(updated['day_of_month'])
        self.assertEqual(updated['next_run'], expected_next_run)

    def test_update_with_day_of_month(self):

        expected_next_run = '1989-01-19T12:00:00Z'

        def fake_schedule_to_next_run(*args, **kwargs):
            return timeutils.parse_isotime(expected_next_run)

        self.stubs.Set(api_utils, 'schedule_to_next_run',
                       fake_schedule_to_next_run)

        request = unit_utils.get_fake_request(method='PUT')
        update_fixture = {'schedule': {'day_of_month': '5'}}

        updated = self.controller.update(request, self.schedule_1['id'],
                                         update_fixture)['schedule']

        self.assertNotEqual(updated.get('created_at'), None)
        self.assertNotEqual(updated.get('updated_at'), None)
        self.assertEqual(self.schedule_1['tenant'], updated['tenant'])
        self.assertEqual(self.schedule_1['action'], updated['action'])
        self.assertEqual(update_fixture['schedule']['day_of_month'],
                         updated['day_of_month'])
        self.assertFalse(updated['minute'])
        self.assertFalse(updated['hour'])
        self.assertFalse(updated['month'])
        self.assertFalse(updated['day_of_week'])
        self.assertEqual(updated['next_run'], expected_next_run)

    def test_update_tenant(self):

        new_tenant = 'new-tenant-name'

        def fake_schedule_to_next_run(*args, **kwargs):
            self.fail('next_run should not be updated')

        self.stubs.Set(api_utils, 'schedule_to_next_run',
                       fake_schedule_to_next_run)

        request = unit_utils.get_fake_request(method='PUT')
        update_fixture = {'schedule': {'tenant': new_tenant}}

        updated = self.controller.update(request, self.schedule_1['id'],
                                         update_fixture)['schedule']

        self.assertNotEqual(updated.get('created_at'), None)
        self.assertNotEqual(updated.get('updated_at'), None)
        self.assertEqual(new_tenant, updated['tenant'])
        self.assertEqual(self.schedule_1['action'], updated['action'])
        self.assertEqual(self.schedule_1['minute'], updated['minute'])
        self.assertEqual(updated['next_run'],
                         timeutils.isotime(self.schedule_1['next_run']))

    def test_update_ignore_read_only(self):
        request = unit_utils.get_fake_request(method='PUT')
        schedule_id = self.schedule_1['id']
        schedule = {'schedule': {'updated_at': '2013-01-02 20:20:20'}}
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.update,
                          request, schedule_id, schedule)

    def test_update_not_found(self):
        request = unit_utils.get_fake_request(method='PUT')
        schedule_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.update,
                          request, schedule_id, {'schedule': {}})

    def test_update_no_body_bad_request(self):
        request = unit_utils.get_fake_request(method='PUT')
        schedule_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          request, schedule_id, None)

    def test_update_no_schedule_bad_request(self):
        request = unit_utils.get_fake_request(method='PUT')
        schedule_id = str(uuid.uuid4())
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          request, schedule_id, {'minute': '5'})



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A filter middleware that inspects the requested URI for a version string
and/or Accept headers and attempts to negotiate an API controller to
return.
"""

from oslo.config import cfg

from qonos.api import versions
from qonos.openstack.common.gettextutils import _
import qonos.openstack.common.log as logging
from qonos.openstack.common import wsgi

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class VersionNegotiationFilter(wsgi.Middleware):

    def __init__(self, app):
        self.versions_app = versions.Controller()
        super(VersionNegotiationFilter, self).__init__(app)

    def process_request(self, req):
        """Try to find a version first in the accept header, then the URL."""
        msg = _("Determining version of request: %(method)s %(path)s"
                " Accept: %(accept)s")
        args = {'method': req.method, 'path': req.path, 'accept': req.accept}
        LOG.debug(msg % args)

        accept = str(req.accept)
        if accept.startswith('application/vnd.openstack.qonos-'):
            LOG.debug(_("Using media-type versioning"))
            token_loc = len('application/vnd.openstack.qonos-')
            req_version = accept[token_loc:]
        else:
            LOG.debug(_("Using url versioning"))
            # Remove version in url so it doesn't conflict later
            req_version = req.path_info_pop()
        try:
            version = self._match_version_string(req_version)
        except ValueError:
            LOG.debug(_("Unknown version. Returning version choices."))
            return self.versions_app

        req.environ['api.version'] = version
        req.path_info = ''.join(('/v', str(version), req.path_info or '/'))
        LOG.debug(_("Matched version: v%d"), version)
        LOG.debug('new uri %s' % req.path_info)
        return None

    def _match_version_string(self, subject):
        """
        Given a string, tries to match a major and/or
        minor version number.

        :param subject: The string to check
        :returns version found in the subject
        :raises ValueError if no acceptable version could be found
        """
        if subject in ('v1', 'v1.0'):
            major_version = 1
        else:
            raise ValueError()

        return major_version

    @classmethod
    def factory(cls, global_conf, **local_conf):
        def filter(app):
            return cls(app)
        return filter



# vim: tabstop=4 shiftwidth=4 softtabstop=4

#    Copyright 2013 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import uuid

from oslo.config import cfg

from qonos.common import exception
from qonos.common import timeutils
from qonos.common import utils as qonos_utils
from qonos.openstack.common import uuidutils
from qonos.tests.unit import utils as unit_utils
from qonos.tests import utils as test_utils


TENANT_1 = uuid.uuid4()
TENANT_2 = uuid.uuid4()


CONF = cfg.CONF


#NOTE(ameade): This is set in each individual db test module
db_api = None


class TestDBApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestDBApi, self).setUp()
        self.db_api = db_api

    def tearDown(self):
        super(TestDBApi, self).setUp()
        self.db_api.reset()

    def test_reset(self):
        fixture = {
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': '30',
            'hour': '2',
        }
        self.db_api.schedule_create(fixture)
        self.db_api.reset()
        self.assertFalse(self.db_api.schedule_get_all())


class TestSchedulesDBApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestSchedulesDBApi, self).setUp()
        self.db_api = db_api
        timeutils.set_time_override(datetime.datetime(
                year=2013, month=1, day=1, hour=9, minute=0))
        self._create_schedules()

    def tearDown(self):
        super(TestSchedulesDBApi, self).setUp()
        self.db_api.reset()
        timeutils.clear_time_override()

    def _create_schedules(self):
        fixture = {
            'id': unit_utils.SCHEDULE_UUID1,
            'tenant': str(TENANT_1),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
            'next_run': qonos_utils.cron_string_to_next_datetime(30, 2),
            'schedule_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance_1',
                },
            ],
        }
        self.schedule_1 = self.db_api.schedule_create(fixture)
        fixture = {
            'id': unit_utils.SCHEDULE_UUID2,
            'tenant': str(TENANT_2),
            'action': 'snapshot',
            'minute': 30,
            'hour': 3,
            'next_run': qonos_utils.cron_string_to_next_datetime(30, 3),
        }
        self.schedule_2 = self.db_api.schedule_create(fixture)

    def _create_basic_schedule(self):
        return db_api.schedule_create({'action': 'snapshot',
                                       'tenant': unit_utils.TENANT1})

    def test_schedule_get_all(self):
        schedules = self.db_api.schedule_get_all()
        self.assertEqual(len(schedules), 2)

    def test_schedule_get_all_filter(self):
        filters = {}
        filters['next_run_after'] = self.schedule_1['next_run']
        filters['next_run_before'] = self.schedule_2['next_run']
        filters['tenant'] = str(TENANT_1)
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 1)
        self.assertEqual(schedules[0]['id'], self.schedule_1['id'])

    def test_schedule_get_all_next_run_filters_are_equal(self):
        filters = {}
        filters['next_run_after'] = self.schedule_1['next_run']
        filters['next_run_before'] = self.schedule_1['next_run']
        filters['tenant'] = str(TENANT_1)
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 1)

    def test_schedule_get_all_tenant_filter(self):
        filters = {}
        filters['tenant'] = str(TENANT_1)
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 1)
        self.assertEqual(schedules[0]['id'], self.schedule_1['id'])

    def test_schedule_get_all_instance_id_filter(self):
        filters = {}
        filters['instance_id'] = 'my_instance_1'
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 1)
        self.assertEqual(schedules[0]['id'], self.schedule_1['id'])

    def test_schedule_get_next_run_filters(self):
        filters = {}
        filters['next_run_after'] = self.schedule_1['next_run']
        filters['next_run_before'] = self.schedule_2['next_run']
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 2)

    def test_schedule_get_all_instance_unknown_filter(self):
        filters = {}
        filters['instance_name'] = 'my_instance_1_name'
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 0)

    def test_schedule_get_next_run_before_filter(self):
        filters = {}
        filters['next_run_before'] = self.schedule_1['next_run']
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 1)

    def test_schedule_get_next_run_after_filter(self):
        filters = {}
        filters['next_run_after'] = self.schedule_2['next_run']
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 1)

    def test_schedule_get_all_with_limit(self):
        filters = {}
        filters['limit'] = 1
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 1)

    def test_schedule_get_all_with_marker(self):
        filters = {}
        filters['marker'] = self.schedule_1['id']
        schedules = self.db_api.schedule_get_all(filter_args=filters)
        self.assertEqual(len(schedules), 1)
        expected = [self.schedule_2]
        self.assertEqual(expected, schedules)

    def test_schedule_get_by_id(self):
        fixture = {
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
            'schedule_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        expected = self.db_api.schedule_create(fixture)
        actual = self.db_api.schedule_get_by_id(expected['id'])
        self.assertEqual(actual['tenant'], fixture['tenant'])
        self.assertEqual(actual['action'], fixture['action'])
        self.assertEqual(actual['minute'], fixture['minute'])
        self.assertEqual(actual['hour'], fixture['hour'])
        metadata = actual['schedule_metadata']
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0]['key'],
                         fixture['schedule_metadata'][0]['key'])
        self.assertEqual(metadata[0]['value'],
                         fixture['schedule_metadata'][0]['value'])
        self.assertNotEqual(actual['created_at'], None)
        self.assertNotEqual(actual['updated_at'], None)

    def test_schedule_get_by_id_not_found(self):
        schedule_id = str(uuid.uuid4())
        self.assertRaises(exception.NotFound,
                          self.db_api.schedule_get_by_id, schedule_id)

    def test_schedule_create_no_action(self):
        fixture = {
            'tenant': unit_utils.TENANT1,
        }
        self.assertRaises(exception.MissingValue,
                          self.db_api.schedule_create, fixture)

    def test_schedule_create_no_tenant(self):
        fixture = {
            'action': 'snapshot',
        }
        self.assertRaises(exception.MissingValue,
                          self.db_api.schedule_create, fixture)

    def test_schedule_create(self):
        fixture = {
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
            'schedule_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        schedule = self.db_api.schedule_create(fixture)
        self.assertTrue(uuidutils.is_uuid_like(schedule['id']))
        self.assertEqual(schedule['tenant'], fixture['tenant'])
        self.assertEqual(schedule['action'], fixture['action'])
        self.assertEqual(schedule['minute'], fixture['minute'])
        self.assertEqual(schedule['hour'], fixture['hour'])
        metadata = schedule['schedule_metadata']
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0]['key'],
                         fixture['schedule_metadata'][0]['key'])
        self.assertEqual(metadata[0]['value'],
                         fixture['schedule_metadata'][0]['value'])
        self.assertNotEqual(schedule['created_at'], None)
        self.assertNotEqual(schedule['updated_at'], None)

    def test_schedule_update(self):
        fixture = {
            'id': str(uuid.uuid4()),
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
        }
        schedule = self.db_api.schedule_create(fixture)
        fixture = {'hour': 3}
        timeutils.set_time_override()
        timeutils.advance_time_seconds(2)
        updated_schedule = self.db_api.schedule_update(schedule['id'], fixture)
        timeutils.clear_time_override()

        self.assertTrue(uuidutils.is_uuid_like(schedule['id']))
        self.assertEqual(updated_schedule['tenant'], schedule['tenant'])
        self.assertEqual(updated_schedule['action'], schedule['action'])
        self.assertEqual(updated_schedule['minute'], schedule['minute'])
        self.assertEqual(updated_schedule['hour'], fixture['hour'])
        self.assertEqual(updated_schedule['created_at'],
                         schedule['created_at'])
        self.assertNotEqual(updated_schedule['updated_at'],
                            schedule['updated_at'])

    def test_schedule_update_remove_metadata(self):
        fixture = {
            'id': str(uuid.uuid4()),
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
            'schedule_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        schedule = self.db_api.schedule_create(fixture)
        fixture = {
            'schedule_metadata': [],
        }

        timeutils.set_time_override()
        timeutils.advance_time_seconds(2)
        updated_schedule = self.db_api.schedule_update(schedule['id'], fixture)
        timeutils.clear_time_override()

        self.assertTrue(uuidutils.is_uuid_like(schedule['id']))
        self.assertEqual(updated_schedule['tenant'], schedule['tenant'])
        self.assertEqual(updated_schedule['action'], schedule['action'])
        self.assertEqual(updated_schedule['minute'], schedule['minute'])
        self.assertEqual(updated_schedule['hour'], schedule['hour'])
        metadata = updated_schedule['schedule_metadata']
        self.assertEqual(len(metadata), 0)
        # updated child metadata collection doesn't update the parent schedule
        self.assertEqual(updated_schedule['updated_at'],
                         schedule['updated_at'])

    def test_schedule_update_metadata(self):
        fixture = {
            'id': str(uuid.uuid4()),
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
        }
        schedule = self.db_api.schedule_create(fixture)
        fixture = {
            'schedule_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }

        timeutils.set_time_override()
        timeutils.advance_time_seconds(2)
        updated_schedule = self.db_api.schedule_update(schedule['id'], fixture)
        timeutils.clear_time_override()

        self.assertTrue(uuidutils.is_uuid_like(schedule['id']))
        self.assertEqual(updated_schedule['tenant'], schedule['tenant'])
        self.assertEqual(updated_schedule['action'], schedule['action'])
        self.assertEqual(updated_schedule['minute'], schedule['minute'])
        self.assertEqual(updated_schedule['hour'], schedule['hour'])
        metadata = updated_schedule['schedule_metadata']
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0]['key'],
                         fixture['schedule_metadata'][0]['key'])
        self.assertEqual(metadata[0]['value'],
                         fixture['schedule_metadata'][0]['value'])
        self.assertEqual(updated_schedule['created_at'],
                         schedule['created_at'])
        # updated child metadata collection doesn't update the parent schedule
        self.assertEqual(updated_schedule['updated_at'],
                         schedule['updated_at'])

    def test_schedule_test_and_set_next_run(self):
        fixture = {
            'id': str(uuid.uuid4()),
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
        }
        new_next_run = timeutils.utcnow()
        schedule = self.db_api.schedule_create(fixture)
        self.db_api.schedule_test_and_set_next_run(
                            schedule['id'], None, new_next_run)

        updated_schedule = self.db_api.schedule_get_by_id(schedule['id'])
        self.assertEqual(updated_schedule['next_run'], new_next_run)

    def test_schedule_test_and_set_next_run_with_expected(self):
        fixture = {
            'id': str(uuid.uuid4()),
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
        }
        new_next_run = timeutils.utcnow()
        schedule = self.db_api.schedule_create(fixture)
        self.db_api.schedule_test_and_set_next_run(
                        schedule['id'], schedule.get('next_run'), new_next_run)

        updated_schedule = self.db_api.schedule_get_by_id(schedule['id'])
        self.assertEqual(updated_schedule['next_run'], new_next_run)

    def test_schedule_test_and_set_next_run_invalid(self):
        fixture = {
            'id': str(uuid.uuid4()),
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
        }
        bad_expected_next_run = timeutils.utcnow()
        timeutils.advance_time_seconds(10)
        schedule = self.db_api.schedule_create(fixture)
        self.assertRaises(exception.NotFound,
                          self.db_api.schedule_test_and_set_next_run,
                          schedule['id'], bad_expected_next_run,
                          timeutils.utcnow())

    def test_schedule_delete(self):
        schedules = self.db_api.schedule_get_all()
        self.assertEqual(len(schedules), 2)
        self.db_api.schedule_delete(self.schedule_1['id'])
        schedules = self.db_api.schedule_get_all()
        self.assertEqual(len(schedules), 1)

    def test_schedule_delete_not_found(self):
        schedule_id = str(uuid.uuid4())
        self.assertRaises(exception.NotFound, self.db_api.schedule_delete,
                          schedule_id)

    def test_medadata_created_with_schedule(self):
        fixture = {
            'tenant': str(uuid.uuid4()),
            'action': 'snapshot',
            'minute': 30,
            'hour': 2,
            'schedule_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        schedule = self.db_api.schedule_create(fixture)
        metadata = self.db_api.schedule_meta_get_all(schedule['id'])
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0]['key'],
                         fixture['schedule_metadata'][0]['key'])
        self.assertEqual(metadata[0]['value'],
                         fixture['schedule_metadata'][0]['value'])

    def test_metadata_create(self):
        schedule = self._create_basic_schedule()
        fixture = {'key': 'key1', 'value': 'value1'}
        meta = db_api.schedule_meta_create(schedule['id'], fixture)
        self.assertEqual(meta['schedule_id'], schedule['id'])
        self.assertEqual(meta['key'], fixture['key'])
        self.assertEqual(meta['value'], fixture['value'])
        self.assertNotEqual(meta['created_at'], None)
        self.assertNotEqual(meta['updated_at'], None)
        self.assertNotEqual(meta['id'], None)

    def test_metadata_create_duplicate(self):
        schedule = self._create_basic_schedule()
        fixture = {'key': 'key1', 'value': 'value1'}
        meta = db_api.schedule_meta_create(schedule['id'], fixture)
        fixture = {'key': 'key1', 'value': 'value1'}

        self.assertRaises(exception.Duplicate, db_api.schedule_meta_create,
                          schedule['id'], fixture)

    def test_metadata_get_all(self):
        schedule = self._create_basic_schedule()
        fixture1 = {'key': 'key1', 'value': 'value1'}
        db_api.schedule_meta_create(schedule['id'], fixture1)
        fixture2 = {'key': 'key2', 'value': 'value2'}
        db_api.schedule_meta_create(schedule['id'], fixture2)
        metadata = db_api.schedule_meta_get_all(schedule['id'])
        self.assertEqual(len(metadata), 2)
        for element in metadata:
            self.assertNotEqual(element['id'], None)
            self.assertEqual(element['schedule_id'], schedule['id'])
            self.assertNotEqual(element['created_at'], None)
            self.assertNotEqual(element['updated_at'], None)
            self.assertEqual(element['schedule_id'], schedule['id'])

        self.assertDbMetaInList(metadata, fixture1)
        self.assertDbMetaInList(metadata, fixture2)

    def test_metadata_get_all_no_meta_create(self):
        schedule = self._create_basic_schedule()
        metadata = db_api.schedule_meta_get_all(schedule['id'])
        self.assertEqual(len(metadata), 0)

    def test_metadata_delete(self):
        schedule = self._create_basic_schedule()
        fixture = {'key': 'key1', 'value': 'value1'}
        meta = db_api.schedule_meta_create(schedule['id'], fixture)
        db_api.schedule_meta_delete(schedule['id'], fixture['key'])
        metadata = db_api.schedule_meta_get_all(schedule['id'])
        self.assertEqual(0, len(metadata))
        self.assertFalse(meta in metadata)

    def test_metadata_delete_not_found(self):
        schedule = self._create_basic_schedule()
        fixture = {'key': 'key1', 'value': 'value1'}
        db_api.schedule_meta_create(schedule['id'], fixture)
        db_api.schedule_meta_delete(schedule['id'], fixture['key'])
        self.assertRaises(exception.NotFound, db_api.schedule_meta_delete,
                          schedule['id'], fixture['key'])

    def test_metadata_update(self):
        schedule = self._create_basic_schedule()
        fixture = [{'key': 'foo', 'value': 'bar'}]
        actual = db_api.schedule_metadata_update(schedule['id'], fixture)

        self.assertEqual(actual[0]['key'], fixture[0]['key'])
        self.assertEqual(actual[0]['value'], fixture[0]['value'])
        self.assertEqual(actual[0]['schedule_id'], schedule['id'])
        self.assertTrue(actual[0]['created_at'])
        self.assertTrue(actual[0]['updated_at'])
        self.assertTrue(actual[0]['id'])

    def test_metadata_update_no_change(self):
        schedule = self._create_basic_schedule()
        fixture = [{'key': 'foo', 'value': 'bar'}]
        actual = db_api.schedule_metadata_update(schedule['id'], fixture)

        fixture = [{'key': 'foo', 'value': 'bar'}]
        actual = db_api.schedule_metadata_update(schedule['id'], fixture)

        self.assertEqual(actual[0]['key'], fixture[0]['key'])
        self.assertEqual(actual[0]['value'], fixture[0]['value'])
        self.assertEqual(actual[0]['schedule_id'], schedule['id'])
        self.assertTrue(actual[0]['created_at'])
        self.assertTrue(actual[0]['updated_at'])
        self.assertTrue(actual[0]['id'])

    def test_metadata_update_delete(self):
        schedule = self._create_basic_schedule()
        fixture = [{'key': 'foo', 'value': 'bar'},
                   {'key': 'foo2', 'value': 'bar2'}]
        db_api.schedule_metadata_update(schedule['id'], fixture)
        actual = db_api.schedule_meta_get_all(schedule['id'])
        self.assertEqual(len(actual), 2)

        fixture = []
        db_api.schedule_metadata_update(schedule['id'], fixture)
        actual = db_api.schedule_meta_get_all(schedule['id'])
        self.assertEqual(actual, [])

    def test_metadata_update_remove_one_metadata_item(self):
        schedule = self._create_basic_schedule()
        fixture = [{'key': 'foo', 'value': 'bar'},
                   {'key': 'foo2', 'value': 'bar2'}]
        db_api.schedule_metadata_update(schedule['id'], fixture)
        original = db_api.schedule_meta_get_all(schedule['id'])
        self.assertEqual(len(original), 2)

        fixture = [{'key': 'foo', 'value': 'bar'}]
        db_api.schedule_metadata_update(schedule['id'], fixture)
        actual = db_api.schedule_meta_get_all(schedule['id'])
        self.assertEqual(len(actual), 1)
        self.assertEqual(original[0]['key'], fixture[0]['key'])
        self.assertEqual(original[0]['value'], fixture[0]['value'])
        self.assertEqual(original[0]['schedule_id'], schedule['id'])
        self.assertTrue(original[0]['created_at'])
        self.assertTrue(original[0]['updated_at'])
        self.assertTrue(original[0]['id'])

    def test_metadata_update_schedule_not_found(self):
        schedule_id = str(uuid.uuid4())
        self.assertRaises(exception.NotFound, db_api.schedule_metadata_update,
                          schedule_id, {})

    def test_metadata_get_all_not_found_when_schedule_doesnt_exists(self):
        schedule_id = str(uuid.uuid4())
        self.assertRaises(exception.NotFound, db_api.schedule_meta_get_all,
                          schedule_id)


class TestWorkersDBApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestWorkersDBApi, self).setUp()
        self.db_api = db_api
        self._create_workers()

    def tearDown(self):
        super(TestWorkersDBApi, self).tearDown()
        self.db_api.reset()

    def _create_workers(self):
        fixture_1 = {'host': 'foo',
                     'id': unit_utils.WORKER_UUID1,
                    }
        fixture_2 = {'host': 'bar',
                     'id': unit_utils.WORKER_UUID2,
                    }
        self.worker_1 = self.db_api.worker_create(fixture_1)
        self.worker_2 = self.db_api.worker_create(fixture_2)

    def test_worker_get_all(self):
        workers = self.db_api.worker_get_all()
        self.assertEqual(len(workers), 2)

    def test_worker_get_all_with_limit(self):
        params = {}
        params['limit'] = 1
        workers = self.db_api.worker_get_all(params=params)
        self.assertEqual(len(workers), 1)
        expected = [self.worker_1]
        self.assertEqual(expected, workers)

    def test_worker_get_all_with_marker(self):
        params = {}
        params['marker'] = self.worker_1['id']
        workers = self.db_api.worker_get_all(params=params)
        self.assertEqual(len(workers), 1)
        expected = [self.worker_2]
        self.assertEqual(expected, workers)

    def test_worker_get_by_id(self):
        actual = self.db_api.worker_get_by_id(self.worker_1['id'])
        self.assertEquals(actual['id'], self.worker_1['id'])
        self.assertEquals(actual['created_at'], self.worker_1['created_at'])
        self.assertEquals(actual['updated_at'], self.worker_1['updated_at'])
        self.assertEquals(actual['host'], self.worker_1['host'])

    def test_worker_get_by_id_not_found(self):
        worker_id = str(uuid.uuid4())
        self.assertRaises(exception.NotFound,
                          self.db_api.worker_get_by_id, worker_id)

    def test_worker_create(self):
        fixture = {'host': 'i.am.cowman'}
        worker = self.db_api.worker_create(fixture)
        self.assertTrue(uuidutils.is_uuid_like(worker['id']))
        self.assertEqual(worker['host'], fixture['host'])
        self.assertNotEqual(worker['created_at'], None)
        self.assertNotEqual(worker['updated_at'], None)

    def test_worker_create_with_pid(self):
        fixture = {'host': 'i.am.cowman',
                   'process_id': 12345}
        worker = self.db_api.worker_create(fixture)
        self.assertTrue(uuidutils.is_uuid_like(worker['id']))
        self.assertEqual(worker['host'], fixture['host'])
        self.assertEqual(worker['process_id'], fixture['process_id'])
        self.assertNotEqual(worker['created_at'], None)
        self.assertNotEqual(worker['updated_at'], None)

    def test_worker_delete(self):
        workers = self.db_api.worker_get_all()
        self.assertEqual(len(workers), 2)
        self.db_api.worker_delete(self.worker_1['id'])
        workers = self.db_api.worker_get_all()
        self.assertEqual(len(workers), 1)

    def test_worker_delete_not_found(self):
        worker_id = str(uuid.uuid4())
        self.assertRaises(exception.NotFound,
                          self.db_api.worker_delete, worker_id)


class TestJobsDBApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestJobsDBApi, self).setUp()
        self.db_api = db_api
        self._create_jobs()

    def tearDown(self):
        super(TestJobsDBApi, self).tearDown()
        self.db_api.reset()

    def _create_jobs(self):
        now = timeutils.utcnow()
        timeout = now + datetime.timedelta(hours=1)
        hard_timeout = now + datetime.timedelta(hours=4)
        fixture = {
            'id': unit_utils.JOB_UUID1,
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID1,
            'worker_id': unit_utils.WORKER_UUID1,
            'status': 'queued',
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 0,
        }
        self.job_1 = self.db_api.job_create(fixture)

        fixture = {
            'id': unit_utils.JOB_UUID2,
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID2,
            'worker_id': unit_utils.WORKER_UUID2,
            'status': 'error',
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 0,
        }
        self.job_2 = self.db_api.job_create(fixture)

    def _create_basic_job(self):
        now = timeutils.utcnow()
        timeout = now + datetime.timedelta(hours=1)
        hard_timeout = now + datetime.timedelta(hours=4)
        return db_api.job_create({
                'action': 'snapshot',
                'timeout': timeout,
                'hard_timeout': hard_timeout,
                'tenant': unit_utils.TENANT1
                })

    def test_job_create_no_action(self):
        fixture = {
            'tenant': unit_utils.TENANT1,
        }
        self.assertRaises(exception.MissingValue,
                          self.db_api.job_create, fixture)

    def test_job_create_no_tenant(self):
        fixture = {
            'action': 'snapshot',
        }
        self.assertRaises(exception.MissingValue,
                          self.db_api.job_create, fixture)

    def test_job_create(self):
        now = timeutils.utcnow()
        timeout = now + datetime.timedelta(hours=1)
        hard_timeout = now + datetime.timedelta(hours=4)
        fixture = {
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID2,
            'worker_id': unit_utils.WORKER_UUID2,
            'status': 'queued',
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'job_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }

        job = self.db_api.job_create(fixture)

        self.assertTrue(uuidutils.is_uuid_like(job['id']))
        self.assertNotEqual(job['created_at'], None)
        self.assertNotEqual(job['updated_at'], None)
        self.assertEqual(job['timeout'], fixture['timeout'])
        self.assertEqual(job['hard_timeout'], fixture['hard_timeout'])
        self.assertEqual(job['schedule_id'], fixture['schedule_id'])
        self.assertEqual(job['worker_id'], fixture['worker_id'])
        self.assertEqual(job['status'], fixture['status'])
        self.assertEqual(job['retry_count'], 0)
        metadata = job['job_metadata']
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0]['key'],
                         fixture['job_metadata'][0]['key'])
        self.assertEqual(metadata[0]['value'],
                         fixture['job_metadata'][0]['value'])

    def test_job_create_no_worker_assigned(self):
        now = timeutils.utcnow()
        timeout = now + datetime.timedelta(hours=1)
        hard_timeout = now + datetime.timedelta(hours=4)
        fixture = {
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID2,
            'status': 'queued',
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'job_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }

        job = self.db_api.job_create(fixture)

        self.assertTrue(uuidutils.is_uuid_like(job['id']))
        self.assertNotEqual(job['created_at'], None)
        self.assertNotEqual(job['updated_at'], None)
        self.assertEqual(job['timeout'], fixture['timeout'])
        self.assertEqual(job['hard_timeout'], fixture['hard_timeout'])
        self.assertEqual(job['schedule_id'], fixture['schedule_id'])
        self.assertEqual(job['worker_id'], None)
        self.assertEqual(job['status'], fixture['status'])
        self.assertEqual(job['retry_count'], 0)
        metadata = job['job_metadata']
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0]['key'],
                         fixture['job_metadata'][0]['key'])
        self.assertEqual(metadata[0]['value'],
                         fixture['job_metadata'][0]['value'])

    def test_jobs_cleanup_hard_timed_out(self):
        jobs = self.db_api.job_get_all()
        self.assertEqual(len(jobs), 2)
        timeutils.set_time_override()
        timeutils.advance_time_delta(datetime.timedelta(hours=4, minutes=1))
        self.db_api._jobs_cleanup_hard_timed_out()
        timeutils.clear_time_override()
        jobs = self.db_api.job_get_all()
        self.assertEqual(len(jobs), 0)

    def test_job_get_all(self):
        jobs = self.db_api.job_get_all()
        self.assertEqual(len(jobs), 2)

    def test_job_get_all_with_limit(self):
        params = {}
        params['limit'] = 1
        jobs = self.db_api.job_get_all(params=params)
        self.assertEqual(len(jobs), 1)
        expected = [self.job_1]
        self.assertEqual(expected, jobs)

    def test_job_get_all_with_marker(self):
        params = {}
        params['marker'] = self.job_1['id']
        jobs = self.db_api.job_get_all(params=params)
        self.assertEqual(len(jobs), 1)
        expected = [self.job_2]
        self.assertEqual(expected, jobs)

    def test_job_get_by_id(self):
        expected = self.job_1
        actual = self.db_api.job_get_by_id(self.job_1['id'])
        self.assertEqual(actual['schedule_id'], expected['schedule_id'])
        self.assertEqual(actual['worker_id'], expected['worker_id'])
        self.assertEqual(actual['status'], expected['status'])
        self.assertEqual(actual['retry_count'], expected['retry_count'])
        self.assertEqual(actual['action'], expected['action'])
        self.assertEqual(actual['tenant'], expected['tenant'])
        self.assertEqual(actual['timeout'], expected['timeout'])
        self.assertEqual(actual['hard_timeout'], expected['hard_timeout'])

    def test_job_get_by_id_not_found(self):
        self.assertRaises(exception.NotFound,
                          self.db_api.job_get_by_id, str(uuid.uuid4))

    def test_job_updated_at_get_by_id(self):
        expected = self.job_1['updated_at']
        actual = self.db_api.job_updated_at_get_by_id(self.job_1['id'])
        self.assertEqual(actual, expected)

    def test_job_updated_at_get_by_id_job_not_found(self):
        self.assertRaises(exception.NotFound,
                          self.db_api.job_updated_at_get_by_id,
                          str(uuid.uuid4))

    def test_job_update(self):
        fixture = {
            'status': 'error',
            'retry_count': 2,
        }
        old = self.db_api.job_get_by_id(self.job_1['id'])
        self.db_api.job_update(self.job_1['id'], fixture)
        updated = self.db_api.job_get_by_id(self.job_1['id'])

        self.assertEqual(old['schedule_id'], updated['schedule_id'])
        self.assertEqual(old['worker_id'], updated['worker_id'])
        self.assertNotEqual(old['status'], updated['status'])
        self.assertNotEqual(old['retry_count'], updated['retry_count'])

        self.assertEqual(updated['status'], 'error')
        self.assertEqual(updated['retry_count'], 2)

    def test_job_update_metadata(self):
        fixture = {
            'job_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        old = self.db_api.job_get_by_id(self.job_1['id'])
        self.db_api.job_update(self.job_1['id'], fixture)
        updated = self.db_api.job_get_by_id(self.job_1['id'])

        self.assertEqual(old['schedule_id'], updated['schedule_id'])
        self.assertEqual(old['worker_id'], updated['worker_id'])
        self.assertEqual(old['status'], updated['status'])
        self.assertEqual(old['retry_count'], updated['retry_count'])

        metadata = updated['job_metadata']
        self.assertEqual(len(old['job_metadata']), 0)
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0]['key'],
                         fixture['job_metadata'][0]['key'])
        self.assertEqual(metadata[0]['value'],
                         fixture['job_metadata'][0]['value'])

    def test_job_update_remove_metadata(self):
        fixture = {
            'job_metadata': [
                {
                    'key': 'instance_id',
                    'value': 'my_instance',
                },
            ],
        }
        old = self.db_api.job_update(self.job_1['id'], fixture)

        fixture = {'job_metadata': []}
        updated = self.db_api.job_update(self.job_1['id'], fixture)

        self.assertEqual(old['schedule_id'], updated['schedule_id'])
        self.assertEqual(old['worker_id'], updated['worker_id'])
        self.assertEqual(old['status'], updated['status'])
        self.assertEqual(old['retry_count'], updated['retry_count'])

        metadata = updated['job_metadata']
        self.assertEqual(len(old['job_metadata']), 1)
        self.assertEqual(len(metadata), 0)

    def test_job_delete(self):
        self.assertEqual(len(self.db_api.job_get_all()), 2)
        self.db_api.job_delete(self.job_1['id'])
        self.assertEqual(len(self.db_api.job_get_all()), 1)

    def test_job_delete_not_found(self):
        self.assertRaises(exception.NotFound,
                          self.db_api.job_delete, str(uuid.uuid4))

    def test_metadata_create(self):
        job = self._create_basic_job()
        fixture = {'key': 'key1', 'value': 'value1'}
        meta = db_api.job_meta_create(job['id'], fixture)
        self.assertEqual(meta['job_id'], job['id'])
        self.assertEqual(meta['key'], fixture['key'])
        self.assertEqual(meta['value'], fixture['value'])
        self.assertNotEqual(meta['created_at'], None)
        self.assertNotEqual(meta['updated_at'], None)
        self.assertNotEqual(meta['id'], None)

    def test_metadata_create_duplicate(self):
        job = self._create_basic_job()
        fixture = {'key': 'key1', 'value': 'value1'}
        meta = db_api.job_meta_create(job['id'], fixture)
        fixture = {'key': 'key1', 'value': 'value1'}

        self.assertRaises(exception.Duplicate, db_api.job_meta_create,
                          job['id'], fixture)

    def test_metadata_get_all_by_job_id(self):
        job = self._create_basic_job()
        fixture = {'key': 'key1', 'value': 'value1'}
        meta = db_api.job_meta_create(job['id'], fixture)
        fixture = {'key': 'key2', 'value': 'value2'}
        meta = db_api.job_meta_create(job['id'], fixture)
        meta_list = db_api.job_meta_get_all_by_job_id(job['id'])
        self.assertEqual(len(meta_list), 2)

    def test_metadata_get_all_by_job_id_no_meta(self):
        job = self._create_basic_job()
        meta_list = db_api.job_meta_get_all_by_job_id(job['id'])
        self.assertEqual(len(meta_list), 0)

    def test_metadata_update(self):
        job = self._create_basic_job()
        fixture = [{'key': 'foo', 'value': 'bar'}]
        actual = db_api.job_metadata_update(job['id'], fixture)

        self.assertEqual(actual[0]['key'], fixture[0]['key'])
        self.assertEqual(actual[0]['value'], fixture[0]['value'])
        self.assertEqual(actual[0]['job_id'], job['id'])
        self.assertTrue(actual[0]['created_at'])
        self.assertTrue(actual[0]['updated_at'])
        self.assertTrue(actual[0]['id'])

    def test_metadata_update_job_not_found(self):
        job_id = str(uuid.uuid4())
        self.assertRaises(exception.NotFound, db_api.job_metadata_update,
                          job_id, {})


class TestJobsDBGetNextJobApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestJobsDBGetNextJobApi, self).setUp()
        self.db_api = db_api
        timeutils.set_time_override()
        self._create_job_fixtures()

    def tearDown(self):
        super(TestJobsDBGetNextJobApi, self).tearDown()
        timeutils.clear_time_override()
        self.db_api.reset()

    def _create_job_fixtures(self):
        now = timeutils.utcnow()
        timeout = now + datetime.timedelta(seconds=30)
        hard_timeout = now + datetime.timedelta(seconds=30)
        self.job_fixture_1 = {
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID1,
            'worker_id': None,
            'status': None,
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 0,
        }

        self.job_fixture_2 = {
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID2,
            'worker_id': unit_utils.WORKER_UUID2,
            'status': 'queued',
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 0,
        }

        self.job_fixture_3 = {
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID1,
            'worker_id': None,
            'status': 'DONE',
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 0,
        }

        self.job_fixture_4 = {
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID1,
            'worker_id': None,
            'status': 'CANCELLED',
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 0,
        }

    def _create_jobs(self, gap, *fixtures):
        now = timeutils.utcnow()
        self.jobs = []
        for fixture in fixtures:
            self.jobs.append(self.db_api.job_create(fixture))
            timeutils.advance_time_seconds(gap)
        return now

    def test_get_next_job_unassigned(self):
        now = timeutils.utcnow()
        new_timeout = now + datetime.timedelta(hours=3)

        retries = 2
        self._create_jobs(10, self.job_fixture_1, self.job_fixture_2)
        job = db_api.job_get_and_assign_next_by_action('snapshot',
                                                       unit_utils.WORKER_UUID1,
                                                       retries,
                                                       new_timeout)
        expected = self.jobs[0]
        self.assertEqual(job['id'], expected['id'])
        self.assertEqual(job['worker_id'], unit_utils.WORKER_UUID1)
        self.assertEqual(job['timeout'], new_timeout)
        self.assertEqual(job['hard_timeout'], expected['hard_timeout'])
        self.assertEqual(job['retry_count'], expected['retry_count'] + 1)

    def test_get_next_job_assigned_once_due_to_timeout(self):
        now = timeutils.utcnow()
        timeout = now - datetime.timedelta(hours=1)
        new_timeout = now + datetime.timedelta(hours=3)
        hard_timeout = now + datetime.timedelta(hours=4)
        job_fixture = {
            'action': 'snapshot',
            'tenant': unit_utils.TENANT1,
            'schedule_id': unit_utils.SCHEDULE_UUID2,
            'worker_id': unit_utils.WORKER_UUID2,
            'status': 'queued',
            'timeout': timeout,
            'hard_timeout': hard_timeout,
            'retry_count': 0,
        }

        retries = 2
        self._create_jobs(10, job_fixture)
        job = db_api.job_get_and_assign_next_by_action('snapshot',
                                                       unit_utils.WORKER_UUID1,
                                                       retries,
                                                       new_timeout)

        job2 = db_api.job_get_and_assign_next_by_action('snapshot',
                                                       unit_utils.WORKER_UUID1,
                                                       retries,
                                                       new_timeout)
        self.assertEqual(job2, None)

    def test_get_next_job_skip_done(self):
        now = timeutils.utcnow()
        new_timeout = now + datetime.timedelta(hours=3)
        retries = 2
        self._create_jobs(10, self.job_fixture_3, self.job_fixture_1)
        job = db_api.job_get_and_assign_next_by_action('snapshot',
                                                       unit_utils.WORKER_UUID1,
                                                       retries,
                                                       new_timeout)
        expected = self.jobs[1]
        self.assertEqual(job['id'], expected['id'])
        self.assertEqual(job['worker_id'], unit_utils.WORKER_UUID1)
        self.assertEqual(job['timeout'], new_timeout)
        self.assertEqual(job['hard_timeout'], expected['hard_timeout'])
        self.assertEqual(job['retry_count'], expected['retry_count'] + 1)

    def test_get_next_job_skip_cancelled(self):
        now = timeutils.utcnow()
        new_timeout = now + datetime.timedelta(hours=3)
        retries = 2
        self._create_jobs(10, self.job_fixture_4, self.job_fixture_1)
        job = db_api.job_get_and_assign_next_by_action('snapshot',
                                                       unit_utils.WORKER_UUID1,
                                                       retries,
                                                       new_timeout)
        expected = self.jobs[1]
        self.assertEqual(job['id'], expected['id'])
        self.assertEqual(job['worker_id'], unit_utils.WORKER_UUID1)
        self.assertEqual(job['timeout'], new_timeout)
        self.assertEqual(job['hard_timeout'], expected['hard_timeout'])
        self.assertEqual(job['retry_count'], expected['retry_count'] + 1)

    def test_get_next_job_timed_out(self):
        now = timeutils.utcnow()
        new_timeout = now + datetime.timedelta(hours=3)
        retries = 2
        self.job_fixture_2['timeout'] = now + datetime.timedelta(seconds=5)
        self._create_jobs(10, self.job_fixture_2, self.job_fixture_1)
        job = db_api.job_get_and_assign_next_by_action('snapshot',
                                                       unit_utils.WORKER_UUID1,
                                                       retries,
                                                       new_timeout)
        expected = self.jobs[0]
        self.assertEqual(job['id'], expected['id'])
        self.assertEqual(job['worker_id'], unit_utils.WORKER_UUID1)
        self.assertEqual(job['timeout'], new_timeout)
        self.assertEqual(job['hard_timeout'], expected['hard_timeout'])
        self.assertEqual(job['retry_count'], expected['retry_count'] + 1)

    def test_get_next_job_too_many_retries(self):
        now = timeutils.utcnow()
        new_timeout = now + datetime.timedelta(hours=3)
        now = timeutils.utcnow()
        retries = 2
        self.job_fixture_2['retry_count'] = 3
        self.job_fixture_2['timeout'] = now + datetime.timedelta(seconds=5)
        self._create_jobs(10, self.job_fixture_2, self.job_fixture_1)
        job = db_api.job_get_and_assign_next_by_action('snapshot',
                                                       unit_utils.WORKER_UUID1,
                                                       retries,
                                                       new_timeout)
        expected = self.jobs[1]
        self.assertEqual(job['id'], expected['id'])
        self.assertEqual(job['worker_id'], unit_utils.WORKER_UUID1)
        self.assertEqual(job['timeout'], new_timeout)
        self.assertEqual(job['hard_timeout'], expected['hard_timeout'])
        self.assertEqual(job['retry_count'], expected['retry_count'] + 1)


class TestJobFaultDBApi(test_utils.BaseTestCase):

    def setUp(self):
        super(TestJobFaultDBApi, self).setUp()
        self.db_api = db_api

    def tearDown(self):
        super(TestJobFaultDBApi, self).tearDown()
        self.db_api.reset()

    def test_job_fault_create_minimal(self):
        fixture = {
            'schedule_id': str(uuid.uuid4()),
            'tenant': str(uuid.uuid4()),
            'worker_id': str(uuid.uuid4()),
            'job_id': str(uuid.uuid4()),
            'action': 'milking',
            }
        job_fault = self.db_api.job_fault_create(fixture)
        self.assertTrue(uuidutils.is_uuid_like(job_fault['id']))
        self.assertEqual(job_fault['schedule_id'], fixture['schedule_id'])
        self.assertEqual(job_fault['tenant'], fixture['tenant'])
        self.assertEqual(job_fault['worker_id'], fixture['worker_id'])
        self.assertEqual(job_fault['job_id'], fixture['job_id'])
        self.assertEqual(job_fault['action'], fixture['action'])
        self.assertNotEqual(job_fault['created_at'], None)
        self.assertNotEqual(job_fault['updated_at'], None)

    def test_job_fault_create_full(self):
        fixture = {
            'schedule_id': str(uuid.uuid4()),
            'tenant': str(uuid.uuid4()),
            'worker_id': str(uuid.uuid4()),
            'job_id': str(uuid.uuid4()),
            'action': 'milking',
            'message': 'There are too many cows here!',
            'job_metadata': 'When were these cows supposed to leave?',
            }
        job_fault = self.db_api.job_fault_create(fixture)
        self.assertTrue(uuidutils.is_uuid_like(job_fault['id']))
        self.assertEqual(job_fault['schedule_id'], fixture['schedule_id'])
        self.assertEqual(job_fault['tenant'], fixture['tenant'])
        self.assertEqual(job_fault['worker_id'], fixture['worker_id'])
        self.assertEqual(job_fault['job_id'], fixture['job_id'])
        self.assertEqual(job_fault['action'], fixture['action'])
        self.assertEqual(job_fault['message'], fixture['message'])
        self.assertEqual(job_fault['job_metadata'], fixture['job_metadata'])
        self.assertNotEqual(job_fault['created_at'], None)
        self.assertNotEqual(job_fault['updated_at'], None)
